import numpy as np
import logging

from rastervision.task import Task
from rastervision.data import ObjectDetectionLabels
from rastervision.core import Box

log = logging.getLogger(__name__)


def _make_chip_pos_windows(image_extent, label_store, chip_size):
    chip_size = chip_size
    pos_windows = []
    boxes = label_store.get_labels().get_boxes()
    done_boxes = set()

    # Get a random window around each box. If a box was previously included
    # in a window, then it is skipped.
    for box in boxes:
        if box.tuple_format() not in done_boxes:
            # If this  object is bigger than the chip,
            # don't use this box.
            if chip_size < box.get_width() or chip_size < box.get_height():
                log.warning('Label is larger than chip size: {} '
                            'Skipping this label'.format(box.tuple_format()))
                continue

            window = box.make_random_square_container(chip_size)
            pos_windows.append(window)

            # Get boxes that lie completely within window
            window_boxes = label_store.get_labels(window=window)
            window_boxes = ObjectDetectionLabels.get_overlapping(
                window_boxes, window, ioa_thresh=1.0)
            window_boxes = window_boxes.get_boxes()
            window_boxes = [box.tuple_format() for box in window_boxes]
            done_boxes.update(window_boxes)

    return pos_windows


def _make_label_pos_windows(image_extent, label_store, label_buffer):
    pos_windows = []
    for box in label_store.get_labels().get_boxes():
        window = box.make_buffer(label_buffer, image_extent)
        pos_windows.append(window)

    return pos_windows


def make_pos_windows(image_extent, label_store, chip_size, window_method,
                     label_buffer):
    if window_method == 'label':
        return _make_label_pos_windows(image_extent, label_store, label_buffer)
    elif window_method == 'image':
        return [image_extent.make_copy()]
    else:
        return _make_chip_pos_windows(image_extent, label_store, chip_size)


def make_neg_windows(raster_source, label_store, chip_size, nb_windows,
                     max_attempts, filter_windows):
    extent = raster_source.get_extent()
    neg_windows = []
    for _ in range(max_attempts):
        for _ in range(max_attempts):
            window = extent.make_random_square(chip_size)
            if any(filter_windows([window])):
                break
        chip = raster_source.get_chip(window)
        labels = ObjectDetectionLabels.get_overlapping(
            label_store.get_labels(), window, ioa_thresh=0.2)

        # If no labels and not blank, append the chip
        if len(labels) == 0 and np.sum(chip.ravel()) > 0:
            neg_windows.append(window)

        if len(neg_windows) == nb_windows:
            break

    return list(neg_windows)


class ObjectDetection(Task):
    def get_train_windows(self, scene):
        raster_source = scene.raster_source
        label_store = scene.ground_truth_label_source

        def filter_windows(windows):
            if scene.aoi_polygons:
                windows = Box.filter_by_aoi(windows, scene.aoi_polygons)
            return windows

        window_method = self.config.chip_options.window_method
        if window_method == 'sliding':
            chip_size = self.config.chip_size
            stride = chip_size
            return list(
                filter_windows((raster_source.get_extent().get_windows(
                    chip_size, stride))))

        # Make positive windows which contain labels.
        pos_windows = filter_windows(
            make_pos_windows(raster_source.get_extent(), label_store,
                             self.config.chip_size,
                             self.config.chip_options.window_method,
                             self.config.chip_options.label_buffer))
        nb_pos_windows = len(pos_windows)

        # Make negative windows which do not contain labels.
        # Generate randow windows and save the ones that don't contain
        # any labels. It may take many attempts to generate a single
        # negative window, and could get into an infinite loop in some cases,
        # so we cap the number of attempts.
        if nb_pos_windows:
            nb_neg_windows = round(
                self.config.chip_options.neg_ratio * nb_pos_windows)
        else:
            nb_neg_windows = 100  # just make some
        max_attempts = 100 * nb_neg_windows
        neg_windows = make_neg_windows(raster_source, label_store,
                                       self.config.chip_size, nb_neg_windows,
                                       max_attempts, filter_windows)

        return pos_windows + neg_windows

    def get_train_labels(self, window, scene):
        window_labels = scene.ground_truth_label_source.get_labels(
            window=window)
        return ObjectDetectionLabels.get_overlapping(
            window_labels,
            window,
            ioa_thresh=self.config.chip_options.ioa_thresh,
            clip=True)

    def get_predict_windows(self, extent):
        chip_size = self.config.chip_size
        stride = chip_size // 2
        return extent.get_windows(chip_size, stride)

    def post_process_predictions(self, labels, scene):
        return ObjectDetectionLabels.prune_duplicates(
            labels,
            score_thresh=self.config.predict_options.score_thresh,
            merge_thresh=self.config.predict_options.merge_thresh)

    def save_debug_predict_image(self, scene, debug_dir_uri):
        # TODO implement this
        pass



import io
from math import ceil

from PIL import Image
import numpy as np
import imageio
import atexit
import logging

log = logging.getLogger(__name__)


def save_img(im_array, output_path):
    imageio.imwrite(output_path, im_array)


def numpy_to_png(array: np.ndarray) -> str:
    """Get a PNG string from a Numpy array.

    Args:
         array: A Numpy array of shape (w, h, 3) or (w, h), where the
               former is meant to become a three-channel image and the
               latter a one-channel image.  The dtype of the array
               should be uint8.

    Returns:
         str

    """
    im = Image.fromarray(array)
    output = io.BytesIO()
    im.save(output, 'png')
    return output.getvalue()


def png_to_numpy(png: str, dtype=np.uint8) -> np.ndarray:
    """Get a Numpy array from a PNG string.

    Args:
         png: A str containing a PNG-formatted image.

    Returns:
         numpy.ndarray

    """
    incoming = io.BytesIO(png)
    im = Image.open(incoming)
    return np.array(im)


def replace_nones_in_dict(target, replace_value):
    """Recursively replaces Nones in a dictionary with the given value."""
    for k in target:
        if target[k] is None:
            target[k] = replace_value
        elif type(target[k]) is list:
            result = []
            for e in target[k]:
                if type(e) is dict:
                    result.append(replace_nones_in_dict(e, replace_value))
                else:
                    if e is None:
                        result.append(replace_value)
                    else:
                        result.append(e)
            target[k] = result
        elif type(target[k]) is dict:
            replace_nones_in_dict(target[k], replace_value)
    return target


def set_nested_keys(target,
                    mods,
                    ignore_missing_keys=False,
                    set_missing_keys=False):
    """Sets dictionary keys based on modifications.

    Args:
       target - Target dictionary to be  modified in-place.
       mods - Dictionary of values to set into the target dict.
              This method will look for any keys matching the mod
              key, even in nested dictionaries. If the mod has a nested
              dictionary, then the leaf key value will only be set
              if that parent dictionary key is found and is a dictionary.
       ignore_missing_keys - If a key is not found, do not throw an error.
       set_missing_keys - If a key is not found, set it. If the key is part
                          of a nested set, and parent keys are found in the target
                          dictionary, then set the key at whatever level of the nested
                          set of keys where the key is first not found.
    """
    searched_keys, found_keys = [], []

    def f(_target, _mods, parent_key=None, mod_parent_key=None):
        for key in _target:
            if key in _mods.keys():
                found_keys.append(key)
                if type(_target[key]) is dict:
                    if type(_mods[key]) is dict:
                        f(_target[key],
                          _mods[key],
                          parent_key=key,
                          mod_parent_key=key)
                    else:
                        raise Exception('Error: cannot modify dict with value')
                else:
                    _target[key] = _mods[key]
            else:
                if type(_target[key]) is dict:
                    f(_target[key],
                      _mods,
                      parent_key=key,
                      mod_parent_key=mod_parent_key)
        searched_keys.extend(list(_mods.keys()))

        if set_missing_keys:
            for key in set(_mods.keys()) - set(found_keys):
                if not type(
                        _mods[key]) is dict and parent_key == mod_parent_key:
                    _target[key] = _mods[key]
                    found_keys.append(key)

    f(target, mods)
    if not ignore_missing_keys:
        d = set(searched_keys) - set(found_keys)
        if d:
            raise Exception('Mod keys not found in target dict: {}'.format(d))


def terminate_at_exit(process):
    def terminate():
        log.debug('Terminating {}...'.format(process.pid))
        process.terminate()

    atexit.register(terminate)


def grouped(lst, size):
    """Returns a list of lists of length 'size'.
    The last list will have size <= 'size'.
    """
    return [lst[n:n + size] for n in range(0, len(lst), size)]


def split_into_groups(lst, num_groups):
    """Attempts to split a list into a given number of groups.
    The number of groups will be at least 1 and at most
    num_groups.

    Args:
       lst:             The list to split
       num_groups:      The number of groups to create.
    Returns:
       A list of size between 1 and num_groups containing lists
       of items of l."""
    group_size = max(int(ceil((len(lst)) / num_groups)), 1)

    return grouped(lst, group_size)



from os.path import join

from torchvision.transforms import Compose, ToTensor
from torch.utils.data import DataLoader

from rastervision.backend.torch_utils.data import DataBunch
from rastervision.backend.torch_utils.chip_classification.folder import (
    ImageFolder)


def build_databunch(data_dir, img_sz, batch_sz, class_names):
    num_workers = 4

    train_dir = join(data_dir, 'train')
    valid_dir = join(data_dir, 'valid')

    aug_transform = Compose([ToTensor()])
    transform = Compose([ToTensor()])

    train_ds = ImageFolder(
        train_dir, transform=aug_transform, classes=class_names)
    valid_ds = ImageFolder(valid_dir, transform=transform, classes=class_names)

    train_dl = DataLoader(
        train_ds,
        shuffle=True,
        batch_size=batch_sz,
        num_workers=num_workers,
        drop_last=True,
        pin_memory=True)
    valid_dl = DataLoader(
        valid_ds,
        batch_size=batch_sz,
        num_workers=num_workers,
        pin_memory=True)

    return DataBunch(train_ds, train_dl, valid_ds, valid_dl, class_names)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: rastervision/protos/experiment.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2
from rastervision.protos import dataset_pb2 as rastervision_dot_protos_dot_dataset__pb2
from rastervision.protos import task_pb2 as rastervision_dot_protos_dot_task__pb2
from rastervision.protos import backend_pb2 as rastervision_dot_protos_dot_backend__pb2
from rastervision.protos import analyzer_pb2 as rastervision_dot_protos_dot_analyzer__pb2
from rastervision.protos import evaluator_pb2 as rastervision_dot_protos_dot_evaluator__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='rastervision/protos/experiment.proto',
  package='rv.protos',
  syntax='proto2',
  serialized_pb=_b('\n$rastervision/protos/experiment.proto\x12\trv.protos\x1a\x1cgoogle/protobuf/struct.proto\x1a!rastervision/protos/dataset.proto\x1a\x1erastervision/protos/task.proto\x1a!rastervision/protos/backend.proto\x1a\"rastervision/protos/analyzer.proto\x1a#rastervision/protos/evaluator.proto\"\xae\x03\n\x10\x45xperimentConfig\x12\n\n\x02id\x18\x01 \x02(\t\x12)\n\x07\x64\x61taset\x18\x02 \x02(\x0b\x32\x18.rv.protos.DatasetConfig\x12#\n\x04task\x18\x03 \x02(\x0b\x32\x15.rv.protos.TaskConfig\x12)\n\x07\x62\x61\x63kend\x18\x04 \x02(\x0b\x32\x18.rv.protos.BackendConfig\x12,\n\tanalyzers\x18\x05 \x03(\x0b\x32\x19.rv.protos.AnalyzerConfig\x12.\n\nevaluators\x18\x06 \x03(\x0b\x32\x1a.rv.protos.EvaluatorConfig\x12\x10\n\x08root_uri\x18\x07 \x02(\t\x12\x13\n\x0b\x61nalyze_uri\x18\x08 \x02(\t\x12\x10\n\x08\x63hip_uri\x18\t \x02(\t\x12\x11\n\ttrain_uri\x18\n \x02(\t\x12\x13\n\x0bpredict_uri\x18\x0b \x02(\t\x12\x10\n\x08\x65val_uri\x18\x0c \x02(\t\x12\x12\n\nbundle_uri\x18\r \x02(\t\x12.\n\rcustom_config\x18\x0e \x01(\x0b\x32\x17.google.protobuf.Struct')
  ,
  dependencies=[google_dot_protobuf_dot_struct__pb2.DESCRIPTOR,rastervision_dot_protos_dot_dataset__pb2.DESCRIPTOR,rastervision_dot_protos_dot_task__pb2.DESCRIPTOR,rastervision_dot_protos_dot_backend__pb2.DESCRIPTOR,rastervision_dot_protos_dot_analyzer__pb2.DESCRIPTOR,rastervision_dot_protos_dot_evaluator__pb2.DESCRIPTOR,])
_sym_db.RegisterFileDescriptor(DESCRIPTOR)




_EXPERIMENTCONFIG = _descriptor.Descriptor(
  name='ExperimentConfig',
  full_name='rv.protos.ExperimentConfig',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='rv.protos.ExperimentConfig.id', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='dataset', full_name='rv.protos.ExperimentConfig.dataset', index=1,
      number=2, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='task', full_name='rv.protos.ExperimentConfig.task', index=2,
      number=3, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='backend', full_name='rv.protos.ExperimentConfig.backend', index=3,
      number=4, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='analyzers', full_name='rv.protos.ExperimentConfig.analyzers', index=4,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='evaluators', full_name='rv.protos.ExperimentConfig.evaluators', index=5,
      number=6, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='root_uri', full_name='rv.protos.ExperimentConfig.root_uri', index=6,
      number=7, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='analyze_uri', full_name='rv.protos.ExperimentConfig.analyze_uri', index=7,
      number=8, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='chip_uri', full_name='rv.protos.ExperimentConfig.chip_uri', index=8,
      number=9, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='train_uri', full_name='rv.protos.ExperimentConfig.train_uri', index=9,
      number=10, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='predict_uri', full_name='rv.protos.ExperimentConfig.predict_uri', index=10,
      number=11, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='eval_uri', full_name='rv.protos.ExperimentConfig.eval_uri', index=11,
      number=12, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='bundle_uri', full_name='rv.protos.ExperimentConfig.bundle_uri', index=12,
      number=13, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='custom_config', full_name='rv.protos.ExperimentConfig.custom_config', index=13,
      number=14, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=257,
  serialized_end=687,
)

_EXPERIMENTCONFIG.fields_by_name['dataset'].message_type = rastervision_dot_protos_dot_dataset__pb2._DATASETCONFIG
_EXPERIMENTCONFIG.fields_by_name['task'].message_type = rastervision_dot_protos_dot_task__pb2._TASKCONFIG
_EXPERIMENTCONFIG.fields_by_name['backend'].message_type = rastervision_dot_protos_dot_backend__pb2._BACKENDCONFIG
_EXPERIMENTCONFIG.fields_by_name['analyzers'].message_type = rastervision_dot_protos_dot_analyzer__pb2._ANALYZERCONFIG
_EXPERIMENTCONFIG.fields_by_name['evaluators'].message_type = rastervision_dot_protos_dot_evaluator__pb2._EVALUATORCONFIG
_EXPERIMENTCONFIG.fields_by_name['custom_config'].message_type = google_dot_protobuf_dot_struct__pb2._STRUCT
DESCRIPTOR.message_types_by_name['ExperimentConfig'] = _EXPERIMENTCONFIG

ExperimentConfig = _reflection.GeneratedProtocolMessageType('ExperimentConfig', (_message.Message,), dict(
  DESCRIPTOR = _EXPERIMENTCONFIG,
  __module__ = 'rastervision.protos.experiment_pb2'
  # @@protoc_insertion_point(class_scope:rv.protos.ExperimentConfig)
  ))
_sym_db.RegisterMessage(ExperimentConfig)


# @@protoc_insertion_point(module_scope)



from abc import abstractmethod
from copy import deepcopy
import numbers

from google.protobuf import (json_format)

import rastervision as rv
from rastervision.core.config import Config, ConfigBuilder
from rastervision.protos.vector_source_pb2 import (VectorSourceConfig as
                                                   VectorSourceConfigMsg)


class VectorSourceConfig(Config):
    def __init__(self,
                 source_type,
                 class_id_to_filter=None,
                 default_class_id=1,
                 line_bufs=None,
                 point_bufs=None):
        self.source_type = source_type
        self.class_id_to_filter = class_id_to_filter
        self.default_class_id = default_class_id
        self.line_bufs = line_bufs
        self.point_bufs = point_bufs

    def has_null_class_bufs(self):
        if self.point_bufs is not None:
            for c, v in self.point_bufs.items():
                if v is None:
                    return True

        if self.line_bufs is not None:
            for c, v in self.line_bufs.items():
                if v is None:
                    return True

        return False

    def to_proto(self):
        msg = VectorSourceConfigMsg(source_type=self.source_type)

        if self.class_id_to_filter is not None:
            # Convert class_ids to str to put into json format.
            class_id_to_filter = dict(
                [(str(class_id), filter)
                 for class_id, filter in self.class_id_to_filter.items()])
            d = {'class_id_to_filter': class_id_to_filter}
            msg.MergeFrom(json_format.ParseDict(d, VectorSourceConfigMsg()))

        if self.default_class_id is not None:
            msg.default_class_id = self.default_class_id

        if self.line_bufs is not None:
            # Convert class_ids to str to put into json format.
            line_bufs = dict([(str(c), v) for c, v in self.line_bufs.items()])
            d = {'line_bufs': line_bufs}
            msg.MergeFrom(json_format.ParseDict(d, VectorSourceConfigMsg()))

        if self.point_bufs is not None:
            # Convert class_ids to str to put into json format.
            point_bufs = dict(
                [(str(c), v) for c, v in self.point_bufs.items()])
            d = {'point_bufs': point_bufs}
            msg.MergeFrom(json_format.ParseDict(d, VectorSourceConfigMsg()))

        return msg

    @staticmethod
    def builder(source_type):
        return rv._registry.get_config_builder(rv.VECTOR_SOURCE, source_type)()

    def to_builder(self):
        return rv._registry.get_config_builder(rv.VECTOR_SOURCE,
                                               self.source_type)(self)

    @staticmethod
    def from_proto(msg):
        """Creates a from the specificed protobuf message.
        """
        return rv._registry.get_config_builder(rv.VECTOR_SOURCE, msg.source_type)() \
                           .from_proto(msg) \
                           .build()

    @abstractmethod
    def create_source(self, crs_transformer=None, extent=None, class_map=None):
        pass


class VectorSourceConfigBuilder(ConfigBuilder):
    def from_proto(self, msg):
        b = self
        d = json_format.MessageToDict(msg)

        class_id_to_filter = None
        if msg.HasField('class_id_to_filter'):
            # Convert class_ids from strs to ints.
            # Have to use camel case after parsing from json :(
            class_id_to_filter = dict(
                [(int(class_id), filter)
                 for class_id, filter in d['classIdToFilter'].items()])

        default_class_id = None
        if msg.HasField('default_class_id'):
            default_class_id = msg.default_class_id

        b = b.with_class_inference(
            class_id_to_filter=class_id_to_filter,
            default_class_id=default_class_id)

        line_bufs = msg.line_bufs if msg.HasField('line_bufs') else None
        point_bufs = msg.point_bufs if msg.HasField('point_bufs') else None
        b = b.with_buffers(line_bufs, point_bufs)

        return b

    def validate(self):
        def _validate(bufs):
            if bufs is not None:
                for c, v in bufs.items():
                    if not (v is None or isinstance(v, numbers.Number)):
                        raise rv.ConfigError(
                            'Buffer size {} must be a number or None.'.format(
                                v))

        _validate(self.config.get('point_bufs'))
        _validate(self.config.get('line_bufs'))
        super().validate()

    def with_class_inference(self, class_id_to_filter=None,
                             default_class_id=1):
        """Set options for inferring the class of each feature.

        For more info on how class inference works, see ClassInference.infer_class()

        Args:
            class_id_to_filter: (dict) map from class_id to JSON filter.
                The filter schema is according to
                https://github.com/mapbox/mapbox-gl-js/blob/c9900db279db776f493ce8b6749966cedc2d6b8a/src/style-spec/feature_filter/index.js  # noqa
            default_class_id: (int) the default class_id to use if class can't be
                inferred using other mechanisms. If a feature defaults to a class_id of
                None, then that feature will be deleted.
        """
        b = deepcopy(self)
        # Ensure class_ids are ints.
        if class_id_to_filter is not None:
            class_id_to_filter = dict(
                [(int(c), f) for c, f in class_id_to_filter.items()])
        b.config['class_id_to_filter'] = class_id_to_filter
        b.config['default_class_id'] = default_class_id
        return b

    def with_buffers(self, line_bufs=None, point_bufs=None):
        """Set options for buffering lines and points into polygons.

        For example, this is useful for buffering lines representing roads so that
        their width roughly matches the width of roads in the imagery.

        Args:
            line_bufs: (dict or None) If none, uses default buffer value of 1.
                Otherwise, a map from class_id to number of pixels to buffer by.
                If the buffer value is None, then no buffering will be performed
                and the LineString or Point won't get converted to a Polygon.
                Not converting to Polygon is incompatible with the currently
                available LabelSources, but may be useful in the future.
            point_bufs: (dict or None) same as above, but used for buffering
                Points into Polygons.
        """
        b = deepcopy(self)

        if line_bufs is not None:
            line_bufs = dict([(int(c), v) for c, v in line_bufs.items()])
        b.config['line_bufs'] = line_bufs

        if point_bufs is not None:
            point_bufs = dict([(int(c), v) for c, v in point_bufs.items()])
        b.config['point_bufs'] = point_bufs

        return b



from copy import deepcopy

import rastervision as rv
from rastervision.command import (CommandConfig, CommandConfigBuilder,
                                  BundleCommand)
from rastervision.protos.command_pb2 \
    import CommandConfig as CommandConfigMsg
from rastervision.rv_config import RVConfig
from rastervision.data import SceneConfig
from rastervision.command.utils import (check_task_type, check_analyzers_type,
                                        check_backend_type)


class BundleCommandConfig(CommandConfig):
    def __init__(self, root_uri, task, backend, scene, analyzers):
        super().__init__(rv.BUNDLE, root_uri)
        self.task = task
        self.backend = backend
        self.scene = scene
        self.analyzers = analyzers

    def create_command(self, tmp_dir=None):
        if not tmp_dir:
            _tmp_dir = RVConfig.get_tmp_dir()
            tmp_dir = _tmp_dir.name
        else:
            _tmp_dir = tmp_dir

        retval = BundleCommand(self)
        retval.set_tmp_dir(_tmp_dir)
        return retval

    def to_proto(self):
        msg = super().to_proto()

        task = self.task.to_proto()
        backend = self.backend.to_proto()
        scene = self.scene.to_proto()
        analyzers = list(map(lambda a: a.to_proto(), self.analyzers))

        b = CommandConfigMsg.BundleConfig(
            task=task, backend=backend, scene=scene, analyzers=analyzers)

        msg.MergeFrom(CommandConfigMsg(bundle_config=b))

        return msg

    def report_io(self):
        io_def = rv.core.CommandIODefinition()
        self.task.report_io(self.command_type, io_def)
        self.backend.report_io(self.command_type, io_def)
        self.scene.report_io(self.command_type, io_def)
        for analyzer in self.analyzers:
            analyzer.report_io(self.command_type, io_def)
        return io_def


class BundleCommandConfigBuilder(CommandConfigBuilder):
    def __init__(self, command_type, prev=None):
        super().__init__(command_type, prev)
        if prev is None:
            self.task = None
            self.backend = None
            self.scene = None
            self.analyzers = None
        else:
            self.task = prev.task
            self.backend = prev.backend
            self.scene = prev.scene
            self.analyzers = prev.analyzers

    def validate(self):
        super().validate()
        if self.task is None:
            raise rv.ConfigError('Task not set for BundleCommandConfig. '
                                 'Use with_task or with_experiment')
        check_task_type(self.task)
        if self.backend is None:
            raise rv.ConfigError('Backend not set for BundleCommandConfig. '
                                 'Use with_backend or with_experiment')
        check_backend_type(self.backend)
        if self.scene is None:
            raise rv.ConfigError(
                'Template scene not set for BundleCommandConfig. '
                'Use with_scene or with_experiment')
        if not isinstance(self.scene, SceneConfig):
            raise rv.ConfigError(
                'Template scene must be of class SceneConfig, got {}'.format(
                    type(self.scene)))
        if self.analyzers is None:
            raise rv.ConfigError('Analyzers not set for BundleCommandConfig. '
                                 'Use with_analyzers or with_experiment')
        check_analyzers_type(self.analyzers)

    def build(self):
        self.validate()
        return BundleCommandConfig(self.root_uri, self.task, self.backend,
                                   self.scene, self.analyzers)

    def from_proto(self, msg):
        b = super().from_proto(msg)

        conf = msg.bundle_config

        task = rv.TaskConfig.from_proto(conf.task)
        backend = rv.BackendConfig.from_proto(conf.backend)
        scene = rv.SceneConfig.from_proto(conf.scene)
        analyzers = list(map(rv.AnalyzerConfig.from_proto, conf.analyzers))

        b = b.with_task(task)
        b = b.with_backend(backend)
        b = b.with_scene(scene)
        b = b.with_analyzers(analyzers)

        return b

    def get_root_uri(self, experiment_config):
        return experiment_config.bundle_uri

    def with_experiment(self, experiment_config):
        b = super().with_experiment(experiment_config)
        b = b.with_task(experiment_config.task)
        b = b.with_backend(experiment_config.backend)
        b = b.with_scene(experiment_config.dataset.all_scenes()[0])
        b = b.with_analyzers(experiment_config.analyzers)
        return b

    def with_task(self, task):
        """Sets a specific task type.

        Args:
            task:  A TaskConfig object.

        """
        b = deepcopy(self)
        b.task = task
        return b

    def with_backend(self, backend):
        b = deepcopy(self)
        b.backend = backend
        return b

    def with_scene(self, scene):
        b = deepcopy(self)
        b.scene = scene
        return b

    def with_analyzers(self, analyzers):
        b = deepcopy(self)
        b.analyzers = analyzers
        return b



from copy import deepcopy
from typing import (List, Dict, Tuple, Union)

import rastervision as rv
from rastervision.task import SemanticSegmentation
from rastervision.core.class_map import (ClassMap, ClassItem)
from rastervision.task import (TaskConfig, TaskConfigBuilder)
from rastervision.protos.task_pb2 import TaskConfig as TaskConfigMsg
from rastervision.protos.class_item_pb2 import ClassItem as ClassItemMsg


class SemanticSegmentationConfig(TaskConfig):
    class ChipOptions:
        def __init__(self,
                     window_method='random_sample',
                     target_classes=None,
                     debug_chip_probability=0.25,
                     negative_survival_probability=1.0,
                     chips_per_scene=1000,
                     target_count_threshold=1000,
                     stride=None):
            self.window_method = window_method
            self.target_classes = target_classes
            self.debug_chip_probability = debug_chip_probability
            self.negative_survival_probability = negative_survival_probability
            self.chips_per_scene = chips_per_scene
            self.target_count_threshold = target_count_threshold
            self.stride = stride

    def __init__(self,
                 class_map,
                 predict_batch_size=10,
                 predict_package_uri=None,
                 debug=True,
                 chip_size=300,
                 predict_chip_size=300,
                 chip_options=None):
        super().__init__(rv.SEMANTIC_SEGMENTATION, predict_batch_size,
                         predict_package_uri, debug)
        self.class_map = class_map
        self.chip_size = chip_size
        self.predict_chip_size = predict_chip_size
        if chip_options is None:
            chip_options = SemanticSegmentationConfig.ChipOptions()
        self.chip_options = chip_options

    def save_bundle_files(self, bundle_dir):
        return (self, [])

    def load_bundle_files(self, bundle_dir):
        return self

    def create_task(self, backend):
        return SemanticSegmentation(self, backend)

    def to_proto(self):
        msg = super().to_proto()
        chip_options = TaskConfigMsg.SemanticSegmentationConfig.ChipOptions(
            window_method=self.chip_options.window_method,
            target_classes=self.chip_options.target_classes,
            debug_chip_probability=self.chip_options.debug_chip_probability,
            negative_survival_probability=self.chip_options.
            negative_survival_probability,
            chips_per_scene=self.chip_options.chips_per_scene,
            target_count_threshold=self.chip_options.target_count_threshold,
            stride=self.chip_options.stride)

        conf = TaskConfigMsg.SemanticSegmentationConfig(
            chip_size=self.chip_size,
            predict_chip_size=self.predict_chip_size,
            class_items=self.class_map.to_proto(),
            chip_options=chip_options)
        msg.MergeFrom(
            TaskConfigMsg(
                semantic_segmentation_config=conf,
                predict_package_uri=self.predict_package_uri))

        return msg


class SemanticSegmentationConfigBuilder(TaskConfigBuilder):
    def __init__(self, prev=None):
        config = {}
        if prev:
            config = {
                'predict_batch_size': prev.predict_batch_size,
                'predict_package_uri': prev.predict_package_uri,
                'debug': prev.debug,
                'class_map': prev.class_map,
                'chip_size': prev.chip_size,
                'predict_chip_size': prev.predict_chip_size,
                'chip_options': prev.chip_options
            }
        super().__init__(SemanticSegmentationConfig, config)

    def from_proto(self, msg):
        conf = msg.semantic_segmentation_config

        negative_survival_probability = conf.chip_options \
                                            .negative_survival_probability

        stride = conf.chip_options.stride
        if stride == 0:
            stride = None

        # for backward compatibility
        predict_chip_size = conf.predict_chip_size
        if predict_chip_size == 0:
            predict_chip_size = conf.chip_size

        return self.with_classes(list(conf.class_items)) \
                .with_predict_batch_size(msg.predict_batch_size) \
                .with_predict_package_uri(msg.predict_package_uri) \
                .with_debug(msg.debug) \
                .with_chip_size(conf.chip_size) \
                .with_predict_chip_size(predict_chip_size) \
                .with_chip_options(
                    window_method=conf.chip_options.window_method,
                    target_classes=list(conf.chip_options.target_classes),
                    debug_chip_probability=conf.chip_options.debug_chip_probability,
                    negative_survival_probability=negative_survival_probability,
                    chips_per_scene=conf.chip_options.chips_per_scene,
                    target_count_threshold=conf.chip_options.target_count_threshold,
                    stride=stride)

    def validate(self):
        super().validate()
        # Segmentation masks are stored as uint8 to save space, so can only handle 256
        # classes. If this is really needed, we can add an option for saving with uint16.
        max_classes = 256
        if len(self.config['class_map']) > max_classes:
            raise rv.ConfigError(
                'Cannot use more than {} classes with semantic segmentation.'.
                format(max_classes))

    def with_classes(
            self, classes: Union[ClassMap, List[str], List[ClassItemMsg], List[
                ClassItem], Dict[str, int], Dict[str, Tuple[int, str]]]):
        """Set the classes for this task.

            Args:
                classes: Either a list of class names, a dict which
                         maps class names to class ids, or a dict
                         which maps class names to a tuple of (class_id, color),
                         where color is a PIL color string.
        """
        b = deepcopy(self)
        b.config['class_map'] = ClassMap.construct_from(classes)
        return b

    def with_chip_size(self, chip_size):
        """Set the chip_size for this task.

        Note that some model implementations have a minimum size of input they
        can handle. A value of > 200 is usually safe.

        Args:
            chip_size: (int) chip size in units of pixels
        """
        b = deepcopy(self)
        b.config['chip_size'] = chip_size
        return b

    def with_predict_chip_size(self, chip_size):
        """Set the chip_size to use only at prediction time for this task.

            Args:
                chip_size: Integer value chip size
        """
        b = deepcopy(self)
        b.config['predict_chip_size'] = chip_size
        return b

    def with_chip_options(self,
                          window_method='random_sample',
                          target_classes=None,
                          debug_chip_probability=0.25,
                          negative_survival_probability=1.0,
                          chips_per_scene=1000,
                          target_count_threshold=1000,
                          stride=None):
        """Sets semantic segmentation configurations for the Chip command.

        Args:
            window_method: Window method to use for chipping. Options are:
                random_sample, sliding
            target_classes: list of class ids to train model on
            debug_chip_probability: probability of generating a debug chip.
                Applies to the 'random_sample' window method.
            negative_survival_probability: probability that a sampled negative
                chip will be utilized if it does not contain more pixels than
                target_count_threshold. Applies to the 'random_sample' window method.
            chips_per_scene: number of chips to generate per scene. Applies to
                the 'random_sample' window method.
            target_count_threshold: minimum number of pixels covering
                target_classes that a chip must have. Applies to the
                'random_sample' window method.
            stride: Stride of windows across image. Defaults to half the chip
                size. Applies to the 'sliding_window' method.

        Returns:
            SemanticSegmentationConfigBuilder
        """
        b = deepcopy(self)

        b.config['chip_options'] = SemanticSegmentationConfig.ChipOptions(
            window_method=window_method,
            target_classes=target_classes,
            debug_chip_probability=debug_chip_probability,
            negative_survival_probability=negative_survival_probability,
            chips_per_scene=chips_per_scene,
            target_count_threshold=target_count_threshold,
            stride=stride)
        return b



from rastervision.data.raster_transformer.raster_transformer \
    import RasterTransformer


class NoopTransformer(RasterTransformer):
    """No-op transformer
    """

    def __init__(self):
        pass

    def transform(self, chip, channel_order=None):
        return chip



class CommandIODefinition:
    """Class which contains a set of inputs and outputs for a command,
       based on the configuration.
    """

    def __init__(self,
                 input_uris=None,
                 output_uris=None,
                 missing_input_messages=None):
        if input_uris is None:
            input_uris = set([])
        if output_uris is None:
            output_uris = set([])
        if missing_input_messages is None:
            missing_input_messages = []

        self.input_uris = input_uris
        self.output_uris = output_uris

        # Messages that declare missing inputs
        self.missing_input_messages = missing_input_messages

    def merge(self, other):
        self.input_uris = self.input_uris.union(other.input_uris)
        self.output_uris = self.output_uris.union(other.output_uris)
        self.missing_input_messages = self.missing_input_messages + \
                                      other.missing_input_messages

    def add_input(self, input_uri):
        self.input_uris.add(input_uri)

    def add_inputs(self, input_uris):
        self.input_uris = self.input_uris.union(set(input_uris))

    def add_output(self, output_uri):
        self.output_uris.add(output_uri)

    def add_outputs(self, output_uris):
        self.output_uris = self.output_uris.union(set(output_uris))

    def add_missing(self, message):
        self.missing_input_messages.append(message)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: rastervision/protos/tf_object_detection/image_resizer.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='rastervision/protos/tf_object_detection/image_resizer.proto',
  package='rastervision.protos.tf_object_detection',
  syntax='proto2',
  serialized_pb=_b('\n;rastervision/protos/tf_object_detection/image_resizer.proto\x12\'rastervision.protos.tf_object_detection\"\xe6\x01\n\x0cImageResizer\x12\x64\n\x19keep_aspect_ratio_resizer\x18\x01 \x01(\x0b\x32?.rastervision.protos.tf_object_detection.KeepAspectRatioResizerH\x00\x12Y\n\x13\x66ixed_shape_resizer\x18\x02 \x01(\x0b\x32:.rastervision.protos.tf_object_detection.FixedShapeResizerH\x00\x42\x15\n\x13image_resizer_oneof\"\x90\x02\n\x16KeepAspectRatioResizer\x12\x1a\n\rmin_dimension\x18\x01 \x01(\x05:\x03\x36\x30\x30\x12\x1b\n\rmax_dimension\x18\x02 \x01(\x05:\x04\x31\x30\x32\x34\x12T\n\rresize_method\x18\x03 \x01(\x0e\x32\x33.rastervision.protos.tf_object_detection.ResizeType:\x08\x42ILINEAR\x12#\n\x14pad_to_max_dimension\x18\x04 \x01(\x08:\x05\x66\x61lse\x12#\n\x14\x63onvert_to_grayscale\x18\x05 \x01(\x08:\x05\x66\x61lse\x12\x1d\n\x15per_channel_pad_value\x18\x06 \x03(\x02\"\xb7\x01\n\x11\x46ixedShapeResizer\x12\x13\n\x06height\x18\x01 \x01(\x05:\x03\x33\x30\x30\x12\x12\n\x05width\x18\x02 \x01(\x05:\x03\x33\x30\x30\x12T\n\rresize_method\x18\x03 \x01(\x0e\x32\x33.rastervision.protos.tf_object_detection.ResizeType:\x08\x42ILINEAR\x12#\n\x14\x63onvert_to_grayscale\x18\x04 \x01(\x08:\x05\x66\x61lse*G\n\nResizeType\x12\x0c\n\x08\x42ILINEAR\x10\x00\x12\x14\n\x10NEAREST_NEIGHBOR\x10\x01\x12\x0b\n\x07\x42ICUBIC\x10\x02\x12\x08\n\x04\x41REA\x10\x03')
)
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

_RESIZETYPE = _descriptor.EnumDescriptor(
  name='ResizeType',
  full_name='rastervision.protos.tf_object_detection.ResizeType',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='BILINEAR', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NEAREST_NEIGHBOR', index=1, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='BICUBIC', index=2, number=2,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='AREA', index=3, number=3,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=798,
  serialized_end=869,
)
_sym_db.RegisterEnumDescriptor(_RESIZETYPE)

ResizeType = enum_type_wrapper.EnumTypeWrapper(_RESIZETYPE)
BILINEAR = 0
NEAREST_NEIGHBOR = 1
BICUBIC = 2
AREA = 3



_IMAGERESIZER = _descriptor.Descriptor(
  name='ImageResizer',
  full_name='rastervision.protos.tf_object_detection.ImageResizer',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='keep_aspect_ratio_resizer', full_name='rastervision.protos.tf_object_detection.ImageResizer.keep_aspect_ratio_resizer', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='fixed_shape_resizer', full_name='rastervision.protos.tf_object_detection.ImageResizer.fixed_shape_resizer', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='image_resizer_oneof', full_name='rastervision.protos.tf_object_detection.ImageResizer.image_resizer_oneof',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=105,
  serialized_end=335,
)


_KEEPASPECTRATIORESIZER = _descriptor.Descriptor(
  name='KeepAspectRatioResizer',
  full_name='rastervision.protos.tf_object_detection.KeepAspectRatioResizer',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='min_dimension', full_name='rastervision.protos.tf_object_detection.KeepAspectRatioResizer.min_dimension', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=600,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='max_dimension', full_name='rastervision.protos.tf_object_detection.KeepAspectRatioResizer.max_dimension', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=1024,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='resize_method', full_name='rastervision.protos.tf_object_detection.KeepAspectRatioResizer.resize_method', index=2,
      number=3, type=14, cpp_type=8, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='pad_to_max_dimension', full_name='rastervision.protos.tf_object_detection.KeepAspectRatioResizer.pad_to_max_dimension', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='convert_to_grayscale', full_name='rastervision.protos.tf_object_detection.KeepAspectRatioResizer.convert_to_grayscale', index=4,
      number=5, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='per_channel_pad_value', full_name='rastervision.protos.tf_object_detection.KeepAspectRatioResizer.per_channel_pad_value', index=5,
      number=6, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=338,
  serialized_end=610,
)


_FIXEDSHAPERESIZER = _descriptor.Descriptor(
  name='FixedShapeResizer',
  full_name='rastervision.protos.tf_object_detection.FixedShapeResizer',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='height', full_name='rastervision.protos.tf_object_detection.FixedShapeResizer.height', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=300,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='width', full_name='rastervision.protos.tf_object_detection.FixedShapeResizer.width', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=300,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='resize_method', full_name='rastervision.protos.tf_object_detection.FixedShapeResizer.resize_method', index=2,
      number=3, type=14, cpp_type=8, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='convert_to_grayscale', full_name='rastervision.protos.tf_object_detection.FixedShapeResizer.convert_to_grayscale', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=613,
  serialized_end=796,
)

_IMAGERESIZER.fields_by_name['keep_aspect_ratio_resizer'].message_type = _KEEPASPECTRATIORESIZER
_IMAGERESIZER.fields_by_name['fixed_shape_resizer'].message_type = _FIXEDSHAPERESIZER
_IMAGERESIZER.oneofs_by_name['image_resizer_oneof'].fields.append(
  _IMAGERESIZER.fields_by_name['keep_aspect_ratio_resizer'])
_IMAGERESIZER.fields_by_name['keep_aspect_ratio_resizer'].containing_oneof = _IMAGERESIZER.oneofs_by_name['image_resizer_oneof']
_IMAGERESIZER.oneofs_by_name['image_resizer_oneof'].fields.append(
  _IMAGERESIZER.fields_by_name['fixed_shape_resizer'])
_IMAGERESIZER.fields_by_name['fixed_shape_resizer'].containing_oneof = _IMAGERESIZER.oneofs_by_name['image_resizer_oneof']
_KEEPASPECTRATIORESIZER.fields_by_name['resize_method'].enum_type = _RESIZETYPE
_FIXEDSHAPERESIZER.fields_by_name['resize_method'].enum_type = _RESIZETYPE
DESCRIPTOR.message_types_by_name['ImageResizer'] = _IMAGERESIZER
DESCRIPTOR.message_types_by_name['KeepAspectRatioResizer'] = _KEEPASPECTRATIORESIZER
DESCRIPTOR.message_types_by_name['FixedShapeResizer'] = _FIXEDSHAPERESIZER
DESCRIPTOR.enum_types_by_name['ResizeType'] = _RESIZETYPE

ImageResizer = _reflection.GeneratedProtocolMessageType('ImageResizer', (_message.Message,), dict(
  DESCRIPTOR = _IMAGERESIZER,
  __module__ = 'rastervision.protos.tf_object_detection.image_resizer_pb2'
  # @@protoc_insertion_point(class_scope:rastervision.protos.tf_object_detection.ImageResizer)
  ))
_sym_db.RegisterMessage(ImageResizer)

KeepAspectRatioResizer = _reflection.GeneratedProtocolMessageType('KeepAspectRatioResizer', (_message.Message,), dict(
  DESCRIPTOR = _KEEPASPECTRATIORESIZER,
  __module__ = 'rastervision.protos.tf_object_detection.image_resizer_pb2'
  # @@protoc_insertion_point(class_scope:rastervision.protos.tf_object_detection.KeepAspectRatioResizer)
  ))
_sym_db.RegisterMessage(KeepAspectRatioResizer)

FixedShapeResizer = _reflection.GeneratedProtocolMessageType('FixedShapeResizer', (_message.Message,), dict(
  DESCRIPTOR = _FIXEDSHAPERESIZER,
  __module__ = 'rastervision.protos.tf_object_detection.image_resizer_pb2'
  # @@protoc_insertion_point(class_scope:rastervision.protos.tf_object_detection.FixedShapeResizer)
  ))
_sym_db.RegisterMessage(FixedShapeResizer)


# @@protoc_insertion_point(module_scope)



# flake8: noqa

from rastervision.augmentor.augmentor import *
from rastervision.augmentor.augmentor_config import *
from rastervision.augmentor.nodata_augmentor import *
from rastervision.augmentor.nodata_augmentor_config import *



import numpy as np

from rastervision.data.label_source.utils import (
    color_to_triple, color_to_integer, rgb_to_int_array)


class SegmentationClassTransformer():
    def __init__(self, class_map):
        color_to_class = dict(
            [(item.color, item.id) for item in class_map.get_items()])

        # color int to class
        color_int_to_class = dict(
            zip([color_to_integer(c) for c in color_to_class.keys()],
                color_to_class.values()))

        def color_int_to_class_fn(color: int) -> int:
            # Convert unspecified colors to class 0 which is "don't care"
            return color_int_to_class.get(color, 0x00)

        self.transform_color_int_to_class = \
            np.vectorize(color_int_to_class_fn, otypes=[np.uint8])

        # class to color triple
        class_to_color_triple = dict(
            zip(color_to_class.values(),
                [color_to_triple(c) for c in color_to_class.keys()]))

        def class_to_channel_color(channel: int, class_id: int) -> int:
            """Given a channel (red, green, or blue) and a class, return the
            intensity of that channel.

            Args:
                 channel: An integer with value 0, 1, or 2
                      representing the channel.
                 class_id: The class id represented as an integer.
            Returns:
                 The intensity of the channel for the color associated
                      with the given class.
            """
            default_triple = (0x00, 0x00, 0x00)
            return class_to_color_triple.get(class_id, default_triple)[channel]

        class_to_r = np.vectorize(
            lambda c: class_to_channel_color(0, c), otypes=[np.uint8])
        class_to_g = np.vectorize(
            lambda c: class_to_channel_color(1, c), otypes=[np.uint8])
        class_to_b = np.vectorize(
            lambda c: class_to_channel_color(2, c), otypes=[np.uint8])
        self.transform_class_to_color = [class_to_r, class_to_g, class_to_b]

    def rgb_to_class(self, rgb_labels):
        color_int_labels = rgb_to_int_array(rgb_labels)
        class_labels = self.transform_color_int_to_class(color_int_labels)
        return class_labels.astype(np.uint8)

    def class_to_rgb(self, class_labels):
        rgb_labels = np.empty(class_labels.shape + (3, ))
        for chan in range(3):
            class_to_channel_color = self.transform_class_to_color[chan]
            rgb_labels[:, :, chan] = class_to_channel_color(class_labels)
        return rgb_labels.astype(np.uint8)



# flake8: noqa

from rastervision.core.box import *
from rastervision.core.class_map import *
from rastervision.core.command_io_definition import *
from rastervision.core.config import *
from rastervision.core.raster_stats import RasterStats
from rastervision.core.training_data import *



"""Raster Vision main program"""
import sys
import os

import click
import logging

import rastervision as rv
from rastervision.experiment import (ExperimentLoader, LoaderError)
from rastervision.runner import (ExperimentRunner)
from rastervision.rv_config import RVConfig

log = logging.getLogger(__name__)


def print_error(msg):
    click.echo(click.style(msg, fg='red'), err=True)


@click.group()
@click.option(
    '--profile', '-p', help='Sets the configuration profile name to use.')
@click.option(
    '-v', '--verbose', help='Sets the output to  be verbose.', count=True)
def main(profile, verbose):
    # Make sure current directory is on PYTHON_PATH
    # so that we can run against modules in current dir.
    sys.path.append(os.curdir)

    # Initialize configuration
    rv._registry.initialize_config(profile=profile, verbosity=verbose + 1)


@main.command(
    'run', short_help='Run Raster Vision commands against Experiments.')
@click.argument('runner')
@click.argument('commands', nargs=-1)
@click.option(
    '--experiment_module',
    '-e',
    help=('Name of an importable module to look for experiment sets '
          'in. If not supplied, experiments will be loaded '
          'from __main__'))
@click.option(
    '--path',
    '-p',
    metavar='PATTERN',
    help=('Path of file containing ExprimentSet to run.'))
@click.option(
    '--dry-run',
    '-n',
    is_flag=True,
    help=('Execute a dry run, which will print out information '
          'about the commands to be run, but will not actually '
          'run the commands'))
@click.option(
    '--skip-file-check',
    '-x',
    is_flag=True,
    help=('Skip the step that verifies that file exist.'))
@click.option(
    '--arg',
    '-a',
    type=(str, str),
    multiple=True,
    metavar='KEY VALUE',
    help=('Pass a parameter to the experiments if the method '
          'parameter list takes in a parameter with that key. '
          'Multiple args can be supplied'))
@click.option(
    '--prefix',
    metavar='PREFIX',
    default='exp_',
    help=('Prefix for methods containing experiments. (default: "exp_")'))
@click.option(
    '--method',
    '-m',
    'methods',
    multiple=True,
    metavar='PATTERN',
    help=('Pattern to match method names to run.'))
@click.option(
    '--filter',
    '-f',
    'filters',
    multiple=True,
    metavar='PATTERN',
    help=('Pattern to match experiment names to run.'))
@click.option(
    '--rerun',
    '-r',
    is_flag=True,
    default=False,
    help=('Rerun commands, regardless if '
          'their output files already exist.'))
@click.option('--tempdir', help=('Temporary directory to use for this run.'))
@click.option(
    '--splits',
    '-s',
    default=1,
    metavar='INTEGER',
    help=('The number of processes to attempt to split each stage into.'))
def run(runner, commands, experiment_module, dry_run, skip_file_check, arg,
        prefix, methods, path, filters, rerun, tempdir, splits):
    """Run Raster Vision commands from experiments, using the
    experiment runner named RUNNER."""

    if tempdir:
        RVConfig.set_tmp_dir(tempdir)

    # Validate runner
    valid_runners = list(
        map(lambda x: x.lower(), rv.ExperimentRunner.list_runners()))
    if runner not in valid_runners:
        print_error('Invalid experiment runner: "{}". '
                    'Must be one of: "{}"'.format(runner,
                                                  '", "'.join(valid_runners)))
        sys.exit(1)

    runner = ExperimentRunner.get_runner(runner)

    if experiment_module and path:
        print_error('Must specify only one of experiment_module or path')
        sys.exit(1)

    if not commands:
        commands = rv.all_commands()
    else:
        commands = list(map(lambda x: x.upper(), commands))

    experiment_args = {}
    for k, v in arg:
        experiment_args[k] = v

    loader = ExperimentLoader(
        experiment_args=experiment_args,
        experiment_method_prefix=prefix,
        experiment_method_patterns=methods,
        experiment_name_patterns=filters)
    try:
        if experiment_module:
            experiments, command_configs = loader.load_from_module(
                experiment_module)
        elif path:
            experiments, command_configs = loader.load_from_file(path)
        else:
            experiments, command_configs = loader.load_from_module('__main__')
    except LoaderError as e:
        print_error(str(e))
        sys.exit(1)

    if not experiments and not commands:
        if experiment_module:
            print_error(
                'No experiments found in {}.'.format(experiment_module))
        elif path:
            print_error('No experiments found in {}.'.format(path))
        else:
            print_error('No experiments found.')

    runner.run(
        experiments,
        command_configs=command_configs,
        commands_to_run=commands,
        rerun_commands=rerun,
        skip_file_check=skip_file_check,
        dry_run=dry_run,
        splits=splits)


@main.command()
@click.option(
    '--experiment-module',
    '-e',
    help=('Name of an importable module to look for experiment sets '
          'in. If not supplied, experiments will be loaded '
          'from __main__'))
@click.option(
    '--arg',
    '-a',
    type=(str, str),
    multiple=True,
    metavar='KEY VALUE',
    help=('Pass a parameter to the experiments if the method '
          'parameter list takes in a parameter with that key. '
          'Multiple args can be supplied'))
def ls(experiment_module, arg):
    """Print out a list of Experiment IDs."""
    if experiment_module:
        module_to_load = experiment_module
    else:
        module_to_load = '__main__'

    experiment_args = {}
    for k, v in arg:
        experiment_args[k] = v

    loader = ExperimentLoader(experiment_args=experiment_args)
    try:
        experiments = loader.load_from_module(module_to_load)
    except LoaderError as e:
        print_error(str(e))
        sys.exit(1)

    if not experiments:
        if experiment_module:
            print_error(
                'No experiments found in {}.'.format(experiment_module))
        else:
            print_error('No experiments found.')

    for e in experiments:
        click.echo('{}'.format(e.id))


# https://stackoverflow.com/questions/48391777/nargs-equivalent-for-options-in-click
class OptionEatAll(click.Option):
    def __init__(self, *args, **kwargs):
        self.save_other_options = kwargs.pop('save_other_options', True)
        nargs = kwargs.pop('nargs', -1)
        assert nargs == -1, 'nargs, if set, must be -1 not {}'.format(nargs)
        super(OptionEatAll, self).__init__(*args, **kwargs)
        self._previous_parser_process = None
        self._eat_all_parser = None

    def add_to_parser(self, parser, ctx):
        def parser_process(value, state):
            value = str(value)
            while state.rargs:
                value = '{} {}'.format(value, state.rargs.pop(0))
            self._previous_parser_process(value, state)

        retval = super(OptionEatAll, self).add_to_parser(parser, ctx)

        for name in self.opts:
            our_parser = parser._long_opt.get(name) or parser._short_opt.get(
                name)
            if our_parser:
                self._eat_all_parser = our_parser
                self._previous_parser_process = our_parser.process
                our_parser.process = parser_process
                break

        return retval


@main.command(
    'predict', short_help='Make predictions using a predict package.')
@click.argument('predict_package')
@click.argument('image_uri')
@click.argument('output_uri')
@click.option(
    '--update-stats',
    '-a',
    is_flag=True,
    help=('Run an analysis on this individual image, as '
          'opposed to using any analysis like statistics '
          'that exist in the prediction package'))
@click.option(
    '--channel-order',
    cls=OptionEatAll,
    help='List of indices comprising channel_order. Example: 2 1 0')
@click.option(
    '--export-config',
    type=click.Path(exists=False),
    help='Exports the configuration to the given output file.')
def predict(predict_package, image_uri, output_uri, update_stats,
            channel_order, export_config):
    """Make predictions on the image at IMAGE_URI
    using PREDICT_PACKAGE and store the
    prediciton output at OUTPUT_URI.
    """
    if channel_order is not None:
        channel_order = [
            int(channel_ind) for channel_ind in channel_order.split(' ')
        ]

    with RVConfig.get_tmp_dir() as tmp_dir:
        predictor = rv.Predictor(predict_package, tmp_dir, update_stats,
                                 channel_order)
        predictor.predict(image_uri, output_uri, export_config)


@main.command(
    'run_command', short_help='Run a command from configuration file.')
@click.argument('command_config_uri')
@click.option('--tempdir')
def run_command(command_config_uri, tempdir):
    """Run a command from a serialized command configuration
    at COMMAND_CONFIG_URI.
    """
    if tempdir is not None:
        RVConfig.set_tmp_dir(tempdir)
    rv.runner.CommandRunner.run(command_config_uri)


if __name__ == '__main__':
    main()



from abc import abstractmethod

import rastervision as rv
from rastervision.core.config import (Config, ConfigBuilder,
                                      BundledConfigMixin)


class RasterTransformerConfig(BundledConfigMixin, Config):
    def __init__(self, transformer_type):
        self.transformer_type = transformer_type

    @abstractmethod
    def create_transformer(self):
        """Create the Transformer that this configuration represents
        """
        pass

    def to_builder(self):
        return rv._registry.get_config_builder(rv.RASTER_TRANSFORMER,
                                               self.transformer_type)(self)

    @staticmethod
    def builder(transformer_type):
        return rv._registry.get_config_builder(rv.RASTER_TRANSFORMER,
                                               transformer_type)()

    @staticmethod
    def from_proto(msg):
        """Creates a TaskConfig from the specificed protobuf message
        """
        return rv._registry.get_config_builder(rv.RASTER_TRANSFORMER,
                                               msg.transformer_type)() \
                           .from_proto(msg) \
                           .build()


class RasterTransformerConfigBuilder(ConfigBuilder):
    pass



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: rastervision/protos/tf_object_detection/box_coder.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from rastervision.protos.tf_object_detection import faster_rcnn_box_coder_pb2 as rastervision_dot_protos_dot_tf__object__detection_dot_faster__rcnn__box__coder__pb2
from rastervision.protos.tf_object_detection import keypoint_box_coder_pb2 as rastervision_dot_protos_dot_tf__object__detection_dot_keypoint__box__coder__pb2
from rastervision.protos.tf_object_detection import mean_stddev_box_coder_pb2 as rastervision_dot_protos_dot_tf__object__detection_dot_mean__stddev__box__coder__pb2
from rastervision.protos.tf_object_detection import square_box_coder_pb2 as rastervision_dot_protos_dot_tf__object__detection_dot_square__box__coder__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='rastervision/protos/tf_object_detection/box_coder.proto',
  package='rastervision.protos.tf_object_detection',
  syntax='proto2',
  serialized_pb=_b('\n7rastervision/protos/tf_object_detection/box_coder.proto\x12\'rastervision.protos.tf_object_detection\x1a\x43rastervision/protos/tf_object_detection/faster_rcnn_box_coder.proto\x1a@rastervision/protos/tf_object_detection/keypoint_box_coder.proto\x1a\x43rastervision/protos/tf_object_detection/mean_stddev_box_coder.proto\x1a>rastervision/protos/tf_object_detection/square_box_coder.proto\"\x87\x03\n\x08\x42oxCoder\x12\\\n\x15\x66\x61ster_rcnn_box_coder\x18\x01 \x01(\x0b\x32;.rastervision.protos.tf_object_detection.FasterRcnnBoxCoderH\x00\x12\\\n\x15mean_stddev_box_coder\x18\x02 \x01(\x0b\x32;.rastervision.protos.tf_object_detection.MeanStddevBoxCoderH\x00\x12S\n\x10square_box_coder\x18\x03 \x01(\x0b\x32\x37.rastervision.protos.tf_object_detection.SquareBoxCoderH\x00\x12W\n\x12keypoint_box_coder\x18\x04 \x01(\x0b\x32\x39.rastervision.protos.tf_object_detection.KeypointBoxCoderH\x00\x42\x11\n\x0f\x62ox_coder_oneof')
  ,
  dependencies=[rastervision_dot_protos_dot_tf__object__detection_dot_faster__rcnn__box__coder__pb2.DESCRIPTOR,rastervision_dot_protos_dot_tf__object__detection_dot_keypoint__box__coder__pb2.DESCRIPTOR,rastervision_dot_protos_dot_tf__object__detection_dot_mean__stddev__box__coder__pb2.DESCRIPTOR,rastervision_dot_protos_dot_tf__object__detection_dot_square__box__coder__pb2.DESCRIPTOR,])
_sym_db.RegisterFileDescriptor(DESCRIPTOR)




_BOXCODER = _descriptor.Descriptor(
  name='BoxCoder',
  full_name='rastervision.protos.tf_object_detection.BoxCoder',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='faster_rcnn_box_coder', full_name='rastervision.protos.tf_object_detection.BoxCoder.faster_rcnn_box_coder', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='mean_stddev_box_coder', full_name='rastervision.protos.tf_object_detection.BoxCoder.mean_stddev_box_coder', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='square_box_coder', full_name='rastervision.protos.tf_object_detection.BoxCoder.square_box_coder', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='keypoint_box_coder', full_name='rastervision.protos.tf_object_detection.BoxCoder.keypoint_box_coder', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='box_coder_oneof', full_name='rastervision.protos.tf_object_detection.BoxCoder.box_coder_oneof',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=369,
  serialized_end=760,
)

_BOXCODER.fields_by_name['faster_rcnn_box_coder'].message_type = rastervision_dot_protos_dot_tf__object__detection_dot_faster__rcnn__box__coder__pb2._FASTERRCNNBOXCODER
_BOXCODER.fields_by_name['mean_stddev_box_coder'].message_type = rastervision_dot_protos_dot_tf__object__detection_dot_mean__stddev__box__coder__pb2._MEANSTDDEVBOXCODER
_BOXCODER.fields_by_name['square_box_coder'].message_type = rastervision_dot_protos_dot_tf__object__detection_dot_square__box__coder__pb2._SQUAREBOXCODER
_BOXCODER.fields_by_name['keypoint_box_coder'].message_type = rastervision_dot_protos_dot_tf__object__detection_dot_keypoint__box__coder__pb2._KEYPOINTBOXCODER
_BOXCODER.oneofs_by_name['box_coder_oneof'].fields.append(
  _BOXCODER.fields_by_name['faster_rcnn_box_coder'])
_BOXCODER.fields_by_name['faster_rcnn_box_coder'].containing_oneof = _BOXCODER.oneofs_by_name['box_coder_oneof']
_BOXCODER.oneofs_by_name['box_coder_oneof'].fields.append(
  _BOXCODER.fields_by_name['mean_stddev_box_coder'])
_BOXCODER.fields_by_name['mean_stddev_box_coder'].containing_oneof = _BOXCODER.oneofs_by_name['box_coder_oneof']
_BOXCODER.oneofs_by_name['box_coder_oneof'].fields.append(
  _BOXCODER.fields_by_name['square_box_coder'])
_BOXCODER.fields_by_name['square_box_coder'].containing_oneof = _BOXCODER.oneofs_by_name['box_coder_oneof']
_BOXCODER.oneofs_by_name['box_coder_oneof'].fields.append(
  _BOXCODER.fields_by_name['keypoint_box_coder'])
_BOXCODER.fields_by_name['keypoint_box_coder'].containing_oneof = _BOXCODER.oneofs_by_name['box_coder_oneof']
DESCRIPTOR.message_types_by_name['BoxCoder'] = _BOXCODER

BoxCoder = _reflection.GeneratedProtocolMessageType('BoxCoder', (_message.Message,), dict(
  DESCRIPTOR = _BOXCODER,
  __module__ = 'rastervision.protos.tf_object_detection.box_coder_pb2'
  # @@protoc_insertion_point(class_scope:rastervision.protos.tf_object_detection.BoxCoder)
  ))
_sym_db.RegisterMessage(BoxCoder)


# @@protoc_insertion_point(module_scope)



import rastervision as rv

import io
import os
import shutil
import tarfile
from os.path import join
from subprocess import (Popen, PIPE, STDOUT)
import glob
import re
import uuid
import logging
from copy import deepcopy

from PIL import Image
import numpy as np
from google.protobuf import text_format, json_format

from rastervision.backend import Backend
from rastervision.data import ObjectDetectionLabels
from rastervision.utils.files import (get_local_path, upload_or_copy, make_dir,
                                      download_if_needed, sync_to_dir,
                                      sync_from_dir, start_sync, list_paths)
from rastervision.utils.misc import (save_img, replace_nones_in_dict,
                                     terminate_at_exit)
from rastervision.rv_config import RVConfig

TRAIN = 'train'
VALIDATION = 'validation'

log = logging.getLogger(__name__)


def save_debug_image(im, labels, class_map, output_path):
    from object_detection.utils import visualization_utils as vis_util

    npboxes = labels.get_npboxes()
    class_ids = labels.get_class_ids()
    scores = labels.get_scores()
    if scores is None:
        scores = [1.0] * len(labels)

    vis_util.visualize_boxes_and_labels_on_image_array(
        im,
        npboxes,
        class_ids,
        scores,
        class_map.get_category_index(),
        use_normalized_coordinates=True,
        line_thickness=2,
        max_boxes_to_draw=None)
    save_img(im, output_path)


def create_tf_example(image, window, labels, class_map, chip_id=''):
    import tensorflow as tf
    from object_detection.utils import dataset_util

    image = Image.fromarray(image)
    image_format = 'png'
    encoded_image = io.BytesIO()
    image.save(encoded_image, format=image_format)
    width, height = image.size

    npboxes = labels.get_npboxes()
    npboxes = ObjectDetectionLabels.global_to_local(npboxes, window)
    npboxes = ObjectDetectionLabels.local_to_normalized(npboxes, window)
    ymins = npboxes[:, 0]
    xmins = npboxes[:, 1]
    ymaxs = npboxes[:, 2]
    xmaxs = npboxes[:, 3]
    class_ids = labels.get_class_ids()
    class_names = [
        class_map.get_by_id(class_id).name.encode('utf8')
        for class_id in class_ids
    ]

    tf_example = tf.train.Example(
        features=tf.train.Features(
            feature={
                'image/height':
                dataset_util.int64_feature(height),
                'image/width':
                dataset_util.int64_feature(width),
                'image/filename':
                dataset_util.bytes_feature(chip_id.encode('utf8')),
                'image/source_id':
                dataset_util.bytes_feature(chip_id.encode('utf8')),
                'image/encoded':
                dataset_util.bytes_feature(encoded_image.getvalue()),
                'image/format':
                dataset_util.bytes_feature(image_format.encode('utf8')),
                'image/object/bbox/xmin':
                dataset_util.float_list_feature(xmins),
                'image/object/bbox/xmax':
                dataset_util.float_list_feature(xmaxs),
                'image/object/bbox/ymin':
                dataset_util.float_list_feature(ymins),
                'image/object/bbox/ymax':
                dataset_util.float_list_feature(ymaxs),
                'image/object/class/text':
                dataset_util.bytes_list_feature(class_names),
                'image/object/class/label':
                dataset_util.int64_list_feature(class_ids)
            }))

    return tf_example


def write_tf_record(tf_examples, output_path: str) -> None:
    """Write an array of TFRecords to the given output path.

    Args:
         tf_examples: An array of TFRecords; a
              list(tensorflow.core.example.example_pb2.Example)
         output_path: The path where the records should be stored.

    Returns:
         None

    """
    import tensorflow as tf

    with tf.python_io.TFRecordWriter(output_path) as writer:
        for tf_example in tf_examples:
            writer.write(tf_example.SerializeToString())


def merge_tf_records(output_path, src_records):
    import tensorflow as tf

    with tf.python_io.TFRecordWriter(output_path) as writer:
        log.info('Merging TFRecords')
        for src_record in src_records:
            for string_record in tf.python_io.tf_record_iterator(src_record):
                writer.write(string_record)


def make_tf_class_map(class_map):
    from rastervision.protos.tf_object_detection.string_int_label_map_pb2 import (
        StringIntLabelMap, StringIntLabelMapItem)

    tf_class_map = StringIntLabelMap()
    tf_items = []
    for class_item in class_map.get_items():
        tf_item = StringIntLabelMapItem(id=class_item.id, name=class_item.name)
        tf_items.append(tf_item)
    tf_class_map.item.extend(tf_items)
    return tf_class_map


def save_tf_class_map(tf_class_map, class_map_path):
    tf_class_map_str = text_format.MessageToString(tf_class_map)
    with open(class_map_path, 'w') as class_map_file:
        class_map_file.write(tf_class_map_str)


def make_tf_examples(training_data, class_map):
    tf_examples = []
    log.info('Creating TFRecord')
    for chip, window, labels in training_data:
        tf_example = create_tf_example(chip, window, labels, class_map)
        tf_examples.append(tf_example)
    return tf_examples


def parse_tfexample(example):
    # Parse image.
    im_str = example.features.feature['image/encoded'].bytes_list.value[0]
    im = Image.open(io.BytesIO(im_str))
    im = np.asarray(im, dtype=np.uint8).copy()

    # Parse labels.
    xmins = example.features.feature['image/object/bbox/xmin'].float_list.value
    ymins = example.features.feature['image/object/bbox/ymin'].float_list.value
    xmaxs = example.features.feature['image/object/bbox/xmax'].float_list.value
    ymaxs = example.features.feature['image/object/bbox/ymax'].float_list.value

    nb_boxes = len(xmins)
    npboxes = np.empty((nb_boxes, 4))
    npboxes[:, 0] = ymins
    npboxes[:, 1] = xmins
    npboxes[:, 2] = ymaxs
    npboxes[:, 3] = xmaxs

    class_ids = example.features.feature[
        'image/object/class/label'].int64_list.value
    class_ids = np.array(class_ids)

    labels = ObjectDetectionLabels(npboxes, class_ids)
    return im, labels


def make_debug_images(record_path, class_map, output_dir):
    import tensorflow as tf

    make_dir(output_dir, check_empty=True)

    log.info('Generating debug chips')
    tfrecord_iter = tf.python_io.tf_record_iterator(record_path)
    for ind, example in enumerate(tfrecord_iter):
        example = tf.train.Example.FromString(example)
        im, labels = parse_tfexample(example)
        # Can't create debug images for non-3band images
        if im.shape[2] != 3:
            log.warning(
                'WARNING: Skipping debug images - Images are not 3 band rasters.'
            )
            return
        output_path = join(output_dir, '{}.png'.format(ind))
        save_debug_image(im, labels, class_map, output_path)


def train(config_path,
          output_dir,
          num_steps,
          model_main_py=None,
          do_monitoring=True):
    output_train_dir = join(output_dir, 'train')
    output_eval_dir = join(output_dir, 'eval')

    model_main_py = model_main_py or '/opt/tf-models/object_detection/model_main.py'

    train_cmd = [
        'python', model_main_py, '--alsologtostderr',
        '--pipeline_config_path={}'.format(config_path),
        '--model_dir={}'.format(output_train_dir),
        '--num_train_steps={}'.format(num_steps),
        '--sample_1_of_n_eval_examples={}'.format(1)
    ]

    log.info('Running train command: {}'.format(' '.join(train_cmd)))

    train_process = Popen(train_cmd, stdout=PIPE, stderr=STDOUT)
    terminate_at_exit(train_process)

    if do_monitoring:
        eval_cmd = [
            'python', model_main_py, '--alsologtostderr',
            '--pipeline_config_path={}'.format(config_path),
            '--checkpoint_dir={}'.format(output_train_dir),
            '--model_dir={}'.format(output_eval_dir)
        ]
        log.info('Running eval command: {}'.format(' '.join(eval_cmd)))

        # Don't let the eval process take up GPU space
        env = deepcopy(os.environ)
        env['CUDA_VISIBLE_DEVICES'] = '-1'
        eval_process = Popen(eval_cmd, env=env)

        tensorboard_process = Popen(
            ['tensorboard', '--logdir={}'.format(output_dir)])
        terminate_at_exit(eval_process)
        terminate_at_exit(tensorboard_process)

    with train_process:
        for line in train_process.stdout:
            log.info(line.decode('utf-8'))

    log.info('-----DONE TRAINING----')
    if do_monitoring:
        eval_process.terminate()
        tensorboard_process.terminate()


def get_last_checkpoint_path(train_root_dir):
    index_paths = glob.glob(join(train_root_dir, 'train', '*.index'))
    checkpoint_ids = []
    for index_path in index_paths:
        match = re.match(r'model.ckpt-(\d+).index',
                         os.path.basename(index_path))
        checkpoint_ids.append(int(match.group(1)))

    if len(checkpoint_ids) == 0:
        return None
    checkpoint_id = max(checkpoint_ids)
    checkpoint_path = join(train_root_dir, 'train',
                           'model.ckpt-{}'.format(checkpoint_id))
    return checkpoint_path


def export_inference_graph(train_root_dir,
                           config_path,
                           output_dir,
                           fine_tune_checkpoint_name,
                           export_py=None):
    export_py = (export_py or
                 '/opt/tf-models/object_detection/export_inference_graph.py')
    checkpoint_path = get_last_checkpoint_path(train_root_dir)
    if checkpoint_path is None:
        log.warning('No checkpoints could be found.')
    else:
        log.info('Exporting checkpoint {}...'.format(checkpoint_path))

        train_process = Popen([
            'python', export_py, '--input_type', 'image_tensor',
            '--pipeline_config_path', config_path,
            '--trained_checkpoint_prefix', checkpoint_path,
            '--output_directory', output_dir
        ])
        train_process.wait()

        inference_graph_path = join(output_dir, 'frozen_inference_graph.pb')

        # Package up the model files for usage as fine tuning checkpoints
        model_checkpoint_files = [
            os.path.join(output_dir, fname) for fname in os.listdir(output_dir)
            if fname.startswith('model.ckpt')
        ]

        with RVConfig.get_tmp_dir() as tmp_dir:
            model_dir = os.path.join(tmp_dir, fine_tune_checkpoint_name)
            make_dir(model_dir)
            model_tar = os.path.join(
                output_dir, '{}.tar.gz'.format(fine_tune_checkpoint_name))
            shutil.copy(inference_graph_path, model_dir)
            for path in model_checkpoint_files:
                shutil.copy(path, model_dir)
            with tarfile.open(model_tar, 'w:gz') as tar:
                tar.add(model_dir, arcname=os.path.basename(model_dir))

        # Move frozen inference graph and clean up generated files.

        output_path = join(output_dir, 'model')
        shutil.move(inference_graph_path, output_path)
        saved_model_dir = join(output_dir, 'saved_model')
        shutil.rmtree(saved_model_dir)


class TrainingPackage(object):
    """Encapsulates the files needed to train a model.

    This encapsulates the files that are generated by the make_chips
    command and that are used by the train command. It takes the URI of the
    directory used to store these files (which is remote or local depending on
    where the command is being executed), generates the URIs of all the
    individual files, and downloads and uploads them. This assumes the
    directory has the following structure:
        label-map.pbtxt
        train-debug-chips.zip
        train.record
        validation-debug-chips.zip
        validation.record
    """

    def __init__(self, base_uri, config, tmp_dir, partition_id):
        """Constructor.

        Creates a temporary directory.

        Args:
            base_uri: (string) URI of directory containing files used to
                train model, possibly remote
            config: TFObjectDetectionConfig
        """

        self.temp_dir = tmp_dir

        self.base_uri = base_uri
        self.base_dir = self.get_local_path(base_uri)

        make_dir(self.base_dir)

        self.config = config

        self.partition_id = partition_id

        self.training_record_uri = join(
            base_uri, 'training',
            'training-{}.record'.format(self.partition_id))
        self.training_local_record_path = join(
            self.base_dir, 'training-{}.record'.format(self.partition_id))
        self.training_download_uri = self.get_local_path(
            join(self.base_uri, 'training'))

        self.validation_record_uri = join(
            base_uri, 'validation',
            'validation-{}.record'.format(self.partition_id))
        self.validation_local_record_path = join(
            self.base_dir, 'validation-{}.record'.format(self.partition_id))
        self.validation_download_uri = self.get_local_path(
            join(self.base_uri, 'validation'))

    def get_local_path(self, uri):
        """Get local version of uri.

        Args:
            uri: (string) URI of file, possibly remote

        Returns:
            (string) path of local version of file
        """
        return get_local_path(uri, self.temp_dir)

    def upload_or_copy(self, uri):
        """Upload file if it's remote.

        This knows how to generate the path to the local copy of the file.

        Args:
            uri: (string) URI of file, possibly remote
        """
        upload_or_copy(self.get_local_path(uri), uri)

    def download_if_needed(self, uri):
        """Download file if it's remote.

        Args:
            uri: (string) URI of file, possibly remote

        Returns:
            (string) path of local file that was downloaded
        """
        return download_if_needed(uri, self.temp_dir)

    def get_debug_chips_uri(self, split):
        """Get URI of debug chips zip file for dataset split.

        Args:
            split: (string) 'train' or 'validation'

        Returns:
            (string) URI of zip file containing debug chips, possibly remote
        """
        return join(self.base_uri, '{}-debug/chips-{}.zip'.format(
            split, self.partition_id))

    def upload(self, debug=False):
        """Upload training and validation data, and class map files.

        Args:
            debug: (bool) if True, also upload the corresponding debug chip
                zip files
        """
        if os.path.exists(self.training_local_record_path):
            upload_or_copy(self.training_local_record_path,
                           self.training_record_uri)
        if os.path.exists(self.validation_local_record_path):
            upload_or_copy(self.validation_local_record_path,
                           self.validation_record_uri)
        if debug:
            if os.path.exists(self.get_debug_chips_uri(TRAIN)):
                self.upload_or_copy(self.get_debug_chips_uri(TRAIN))
            if os.path.exists(self.get_debug_chips_uri(VALIDATION)):
                self.upload_or_copy(self.get_debug_chips_uri(VALIDATION))

    def download_data(self):
        """Download training and validation data, and class map files."""

        def _download(split, output_dir):
            for uri in list_paths(self.base_uri, 'record'):
                base_name = os.path.basename(uri)
                if base_name.startswith(split):
                    record_path = self.download_if_needed(uri)
                    target_record_path = os.path.join(
                        output_dir, os.path.basename(record_path))
                    shutil.move(record_path, target_record_path)

        _download('training', self.training_download_uri)
        _download('validation', self.validation_download_uri)

    def download_pretrained_model(self, pretrained_model_zip_uri):
        """Download pretrained model and unzip it.

        This is used before training a model.

        Args:
            pretrained_model_zip_uri: (string) URI of .tar.gz file containing
                pretrained model. This file is of the form that comes from the
                Model Zoo at https://github.com/tensorflow/models/blob/master/research/object_detection/g3doc/detection_model_zoo.md # noqa
                It contains a directory at the top level with the same name as
                root of the zip file (if zip file is x.tar.gz, the directory
                is x), and a set of files of the form model.ckpt.*. This file
                can be located anywhere, and is not expected to be in the
                directory encapsulated by this class that is generated by the
                make_chips command. That is why it is passed in
                separately.

        Returns:
            (string) path to pretrained model file (which is model.ckpt in
                the zip file)
        """
        pretrained_model_zip_path = self.download_if_needed(
            pretrained_model_zip_uri)
        pretrained_model_dir = join(self.temp_dir, 'pretrained_model')
        make_dir(pretrained_model_dir)
        with tarfile.open(pretrained_model_zip_path, 'r:gz') as tar:
            tar.extractall(pretrained_model_dir)
        model_name = os.path.splitext(
            os.path.splitext(os.path.basename(pretrained_model_zip_uri))[0])[0]
        # The unzipped file is assumed to have a single directory with
        # the name of the model derived from the zip file.
        pretrained_model_path = join(pretrained_model_dir, model_name,
                                     'model.ckpt')
        return pretrained_model_path

    def download_config(self, class_map):
        from rastervision.protos.tf_object_detection.pipeline_pb2 \
            import TrainEvalPipelineConfig
        """Download a model and backend config and update its fields.

        This is used before training a model. This downloads and unzips a bunch
        of files that are needed to train a model, and then downloads and
        updates the backend config file with local paths to these files. These
        files include the pretrained model, the class map, and the training and
        validation datasets.

        Args:
            pretrained_model_zip_uri: (string) URI of .tar.gz file containing
                pretrained model. (See download_pretrained_model method for more
                details.)
            backend_config_uri: (string) URI of backend config file which is
                a config file for the TF Object Detection API. Examples can be
                found here https://github.com/tensorflow/models/tree/master/research/object_detection/samples/configs  # noqa
        """

        # Parse configuration
        # We must remove 'nulls' that appear due to translating empty
        # messages. These appear when translating between text and JSON based
        # protobuf messages, and using the google.protobuf.Struct type to store
        # the JSON. This appears when TFOD uses empty message types as an enum.
        config = json_format.ParseDict(
            replace_nones_in_dict(self.config.tfod_config, {}),
            TrainEvalPipelineConfig())

        # Update config using local paths.
        if config.train_config.fine_tune_checkpoint:
            pretrained_model_path = self.download_pretrained_model(
                config.train_config.fine_tune_checkpoint)
            config.train_config.fine_tune_checkpoint = pretrained_model_path

        # Save TF label map based on class_map.
        class_map_path = os.path.join(self.temp_dir, 'label-map.pbtxt')
        tf_class_map = make_tf_class_map(class_map)
        save_tf_class_map(tf_class_map, class_map_path)

        train_record_uris = list_paths(self.training_download_uri, 'record')
        config.train_input_reader.tf_record_input_reader.input_path[:] = train_record_uris
        config.train_input_reader.label_map_path = class_map_path

        eval_record_uris = list_paths(self.validation_download_uri, 'record')
        config.eval_input_reader[
            0].tf_record_input_reader.input_path[:] = eval_record_uris
        config.eval_input_reader[0].label_map_path = class_map_path

        # Save an updated copy of the config file.
        config_path = join(self.temp_dir, 'ml.config')
        config_str = text_format.MessageToString(config)
        with open(config_path, 'w') as config_file:
            config_file.write(config_str)
        return config_path


def load_frozen_graph(inference_graph_path):
    import tensorflow as tf

    detection_graph = tf.Graph()
    with detection_graph.as_default():
        od_graph_def = tf.GraphDef()
        with tf.gfile.GFile(inference_graph_path, 'rb') as fid:
            serialized_graph = fid.read()
            od_graph_def.ParseFromString(serialized_graph)
            tf.import_graph_def(od_graph_def, name='')
    return detection_graph


def compute_prediction(image_nps, windows, detection_graph, session):
    image_tensor = detection_graph.get_tensor_by_name('image_tensor:0')
    boxes = detection_graph.get_tensor_by_name('detection_boxes:0')
    scores = detection_graph.get_tensor_by_name('detection_scores:0')
    class_ids = detection_graph.get_tensor_by_name('detection_classes:0')

    (boxes, scores, class_ids) = session.run(
        [boxes, scores, class_ids], feed_dict={image_tensor: image_nps})

    labels = ObjectDetectionLabels.make_empty()
    for chip_boxes, chip_scores, chip_class_ids, window in zip(
            boxes, scores, class_ids, windows):
        chip_boxes = ObjectDetectionLabels.normalized_to_local(
            chip_boxes, window)
        chip_boxes = ObjectDetectionLabels.local_to_global(chip_boxes, window)
        chip_class_ids = chip_class_ids.astype(np.int32)
        labels = (labels + ObjectDetectionLabels(
            chip_boxes, chip_class_ids, scores=chip_scores))

    return labels


class TFObjectDetection(Backend):
    def __init__(self, backend_config, task_config):
        self.detection_graph = None
        # persist scene training packages for when output_uri is remote
        self.scene_training_packages = []
        self.config = backend_config
        self.class_map = task_config.class_map

        self.partition_id = uuid.uuid4()

    def process_scene_data(self, scene, data, tmp_dir):
        """Process each scene's training data

        Args:
            scene: Scene
            data: TrainingData
            class_map: ClassMap

        Returns:
            the local path to the scene's TFRecord
        """
        # Currently TF Object Detection can only handle uint8
        if scene.raster_source.get_dtype() != np.uint8:
            raise Exception('Cannot use {} backend for imagery that does '
                            'not have data type uint8. '
                            'Use the StatsAnalyzer and StatsTransformer '
                            'to turn the raster data into uint8 data'.format(
                                rv.TF_OBJECT_DETECTION))

        training_package = TrainingPackage(self.config.training_data_uri,
                                           self.config, tmp_dir,
                                           self.partition_id)
        self.scene_training_packages.append(training_package)
        tf_examples = make_tf_examples(data, self.class_map)
        # Ensure directory is unique since scene id's could be shared between
        # training and test sets.
        record_path = os.path.join(
            tmp_dir, '{}-{}.record'.format(scene.id, uuid.uuid4()))
        log.debug('Writing to record path {}'.format(record_path))
        write_tf_record(tf_examples, record_path)

        return record_path

    def process_sceneset_results(self, training_results, validation_results,
                                 tmp_dir):
        """After all scenes have been processed, merge all TFRecords

        Args:
            training_results: list of training scenes' TFRecords
            validation_results: list of validation scenes' TFRecords
        """
        training_package = TrainingPackage(self.config.training_data_uri,
                                           self.config, tmp_dir,
                                           self.partition_id)

        def _merge_training_results(results, record_path, split):
            # merge each scene's tfrecord into "split" tf record
            merge_tf_records(record_path, results)

            # Save debug chips.
            if self.config.debug:
                debug_zip_path = training_package.get_local_path(
                    training_package.get_debug_chips_uri(split))
                with RVConfig.get_tmp_dir() as debug_dir:
                    make_debug_images(record_path, self.class_map, debug_dir)
                    shutil.make_archive(
                        os.path.splitext(debug_zip_path)[0], 'zip', debug_dir)

        if training_results:
            _merge_training_results(
                training_results, training_package.training_local_record_path,
                TRAIN)
        else:
            log.warn('No training chips for partition {}'.format(
                training_package.partition_id))

        if validation_results:
            _merge_training_results(
                validation_results,
                training_package.validation_local_record_path, VALIDATION)
        else:
            log.warn('No validation chips for partition {}'.format(
                training_package.partition_id))

        training_package.upload(debug=self.config.debug)

        # clear scene training packages
        del self.scene_training_packages[:]

    def train(self, tmp_dir):
        training_package = TrainingPackage(self.config.training_data_uri,
                                           self.config, tmp_dir,
                                           self.partition_id)
        # Download training data and update config file.
        training_package.download_data()
        config_path = training_package.download_config(self.class_map)

        # Setup output dirs.
        output_dir = get_local_path(self.config.training_output_uri, tmp_dir)
        make_dir(output_dir)

        # Get output from potential previous run so we can resume training.
        if not self.config.train_options.replace_model:
            sync_from_dir(self.config.training_output_uri, output_dir)
        else:
            for f in os.listdir(output_dir):
                if not f.startswith('command-config'):
                    path = os.path.join(output_dir, f)
                    if os.path.isfile(path):
                        os.remove(path)
                    else:
                        shutil.rmtree(path)

        local_config_path = os.path.join(output_dir, 'pipeline.config')
        shutil.copy(config_path, local_config_path)

        model_main_py = self.config.script_locations.model_main_uri
        export_py = self.config.script_locations.export_uri

        # Train model and sync output periodically.
        sync = start_sync(
            output_dir,
            self.config.training_output_uri,
            sync_interval=self.config.train_options.sync_interval)
        with sync:
            train(
                local_config_path,
                output_dir,
                self.config.get_num_steps(),
                model_main_py=model_main_py,
                do_monitoring=self.config.train_options.do_monitoring)

        export_inference_graph(
            output_dir,
            local_config_path,
            output_dir,
            fine_tune_checkpoint_name=self.config.fine_tune_checkpoint_name,
            export_py=export_py)

        # Perform final sync
        sync_to_dir(output_dir, self.config.training_output_uri)

    def load_model(self, tmp_dir):
        import tensorflow as tf

        # Load and memoize the detection graph and TF session.
        if self.detection_graph is None:
            model_path = download_if_needed(self.config.model_uri, tmp_dir)
            self.detection_graph = load_frozen_graph(model_path)
            self.session = tf.Session(graph=self.detection_graph)

    def predict(self, chips, windows, tmp_dir):
        # Ensure model is loaded
        self.load_model(tmp_dir)

        return compute_prediction(chips, windows, self.detection_graph,
                                  self.session)



import math
import random

import numpy as np
from shapely.geometry import box as ShapelyBox


class BoxSizeError(ValueError):
    pass


class Box():
    """A multi-purpose box (ie. rectangle)."""

    def __init__(self, ymin, xmin, ymax, xmax):
        """Construct a bounding box.

        Unless otherwise stated, the convention is that these coordinates are
        in pixel coordinates and represent boxes that lie within a
        RasterSource.

        Args:
            ymin: minimum y value (y is row)
            xmin: minimum x value (x is column)
            ymax: maximum y value
            xmax: maximum x value

        """
        self.ymin = ymin
        self.xmin = xmin
        self.ymax = ymax
        self.xmax = xmax

    def __eq__(self, other):
        """Return true if other has same coordinates."""
        return self.tuple_format() == other.tuple_format()

    def __ne__(self, other):
        """Return true if other has different coordinates."""
        return self.tuple_format() != other.tuple_format()

    def get_height(self):
        """Return height of Box."""
        return self.ymax - self.ymin

    def get_width(self):
        """Return width of Box."""
        return self.xmax - self.xmin

    def get_area(self):
        """Return area of Box."""
        return self.get_height() * self.get_width()

    def rasterio_format(self):
        """Return Box in Rasterio format."""
        return ((self.ymin, self.ymax), (self.xmin, self.xmax))

    def tuple_format(self):
        return (self.ymin, self.xmin, self.ymax, self.xmax)

    def shapely_format(self):
        return (self.xmin, self.ymin, self.xmax, self.ymax)

    def to_int(self):
        return Box(
            int(self.ymin), int(self.xmin), int(self.ymax), int(self.xmax))

    def npbox_format(self):
        """Return Box in npbox format used by TF Object Detection API.

        Returns:
            Numpy array of form [ymin, xmin, ymax, xmax] with float type

        """
        return np.array(
            [self.ymin, self.xmin, self.ymax, self.xmax], dtype=np.float)

    @staticmethod
    def to_npboxes(boxes):
        """Return nx4 numpy array from list of Box."""
        nb_boxes = len(boxes)
        npboxes = np.empty((nb_boxes, 4))
        for boxind, box in enumerate(boxes):
            npboxes[boxind, :] = box.npbox_format()
        return npboxes

    def __str__(self):  # pragma: no cover
        return str(self.npbox_format())

    def __repr__(self):  # pragma: no cover
        return str(self)

    def geojson_coordinates(self):
        """Return Box as GeoJSON coordinates."""
        # Compass directions:
        nw = [self.xmin, self.ymin]
        ne = [self.xmin, self.ymax]
        se = [self.xmax, self.ymax]
        sw = [self.xmax, self.ymin]
        return [nw, ne, se, sw, nw]

    def make_random_square_container(self, size):
        """Return a new square Box that contains this Box.

        Args:
            size: the width and height of the new Box

        """
        if size < self.get_width():
            raise BoxSizeError('size of random container cannot be < width')

        if size < self.get_height():  # pragma: no cover
            raise BoxSizeError('size of random container cannot be < height')

        lb = self.ymin - (size - self.get_height())
        ub = self.ymin
        rand_y = random.randint(lb, ub)

        lb = self.xmin - (size - self.get_width())
        ub = self.xmin
        rand_x = random.randint(lb, ub)

        return Box.make_square(rand_y, rand_x, size)

    def make_random_square(self, size):
        """Return new randomly positioned square Box that lies inside this Box.

        Args:
            size: the height and width of the new Box

        """
        if size >= self.get_width():
            raise BoxSizeError('size of random square cannot be >= width')

        if size >= self.get_height():  # pragma: no cover
            raise BoxSizeError('size of random square cannot be >= height')

        lb = self.ymin
        ub = self.ymax - size
        rand_y = random.randint(lb, ub)

        lb = self.xmin
        ub = self.xmax - size
        rand_x = random.randint(lb, ub)

        return Box.make_square(rand_y, rand_x, size)

    def intersection(self, other):
        """Return the intersection of this Box and the other.

        Args:
            other: The box to intersect with this one.

        Returns:
             The intersection of this box and the other one.

        """
        xmin = max(self.xmin, other.xmin)
        ymin = max(self.ymin, other.ymin)
        xmax = min(self.xmax, other.xmax)
        ymax = min(self.ymax, other.ymax)
        return Box(xmin=xmin, ymin=ymin, xmax=xmax, ymax=ymax)

    @staticmethod
    def from_npbox(npbox):
        """Return new Box based on npbox format.

        Args:
            npbox: Numpy array of form [ymin, xmin, ymax, xmax] with float type

        """
        return Box(*npbox)

    @staticmethod
    def from_shapely(shape):
        bounds = shape.bounds
        return Box(bounds[1], bounds[0], bounds[3], bounds[2])

    @staticmethod
    def from_tuple(tup):
        """Return new Box based on tuple format.

        Args:
           tup: Tuple format box (ymin, xmin, ymax, xmax)
        """
        return Box(tup[0], tup[1], tup[2], tup[3])

    def to_shapely(self):
        return ShapelyBox(*(self.shapely_format()))

    def reproject(self, transform_fn):
        """Reprojects this box based on a transform function.

        Args:
          transform_fn - A function that takes in a tuple (x, y)
                         and reprojects that point to the target
                         coordinate reference system.
        """
        (xmin, ymin) = transform_fn((self.xmin, self.ymin))
        (xmax, ymax) = transform_fn((self.xmax, self.ymax))

        return Box(ymin, xmin, ymax, xmax)

    @staticmethod
    def make_square(ymin, xmin, size):
        """Return new square Box."""
        return Box(ymin, xmin, ymin + size, xmin + size)

    def make_eroded(self, erosion_size):
        """Return new Box whose sides are eroded by erosion_size."""
        return Box(self.ymin + erosion_size, self.xmin + erosion_size,
                   self.ymax - erosion_size, self.xmax - erosion_size)

    def make_buffer(self, buffer_size, max_extent):
        """Return new Box whose sides are buffered by buffer_size.

        The resulting box is clipped so that the values of the corners are
        always greater than zero and less than the height and width of
        max_extent.

        """
        buffer_size = max(0., buffer_size)
        if buffer_size < 1.:
            delta_width = int(round(buffer_size * self.get_width()))
            delta_height = int(round(buffer_size * self.get_height()))
        else:
            delta_height = delta_width = int(round(buffer_size))

        return Box(
            max(0, math.floor(self.ymin - delta_height)),
            max(0, math.floor(self.xmin - delta_width)),
            min(max_extent.get_height(),
                int(self.ymax) + delta_height),
            min(max_extent.get_width(),
                int(self.xmax) + delta_width))

    def make_copy(self):
        return Box(*(self.tuple_format()))

    def get_windows(self, chip_size, stride):
        """Return list of grid of boxes within this box.

        Args:
            chip_size: (int) the length of each square-shaped window in pixels
            stride: (int) how much each window is offset from the last in pixels

        """
        height = self.get_height()
        width = self.get_width()

        result = []
        for row_start in range(0, height, stride):
            for col_start in range(0, width, stride):
                result.append(Box.make_square(row_start, col_start, chip_size))
        return result

    def to_dict(self):
        return {
            'xmin': self.xmin,
            'ymin': self.ymin,
            'xmax': self.xmax,
            'ymax': self.ymax
        }

    @classmethod
    def from_dict(cls, d):
        return cls(d['ymin'], d['xmin'], d['ymax'], d['xmax'])

    @staticmethod
    def filter_by_aoi(windows, aoi_polygons):
        """Filters windows by a list of AOI polygons"""
        result = []
        for window in windows:
            w = window.to_shapely()
            for polygon in aoi_polygons:
                if w.within(polygon):
                    result.append(window)
                    break

        return result



import keras
from rastervision.protos.keras_classification.optimizer_pb2 import Optimizer


def build(optimizer_options):
    if optimizer_options.type == Optimizer.Type.Value('ADAM'):
        optimizer = keras.optimizers.Adam(lr=optimizer_options.init_lr)
    elif optimizer_options.type == Optimizer.Type.Value('SGD'):
        optimizer = keras.optimizers.SGD(lr=optimizer_options.init_lr)
    else:
        raise ValueError((Optimizer.Type.Name(optimizer_options.type) +
                          ' is not a valid optimizer type'))

    return optimizer



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: rastervision/protos/raster_source.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2
from rastervision.protos import raster_transformer_pb2 as rastervision_dot_protos_dot_raster__transformer__pb2
from rastervision.protos import vector_source_pb2 as rastervision_dot_protos_dot_vector__source__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='rastervision/protos/raster_source.proto',
  package='rv.protos',
  syntax='proto2',
  serialized_pb=_b('\n\'rastervision/protos/raster_source.proto\x12\trv.protos\x1a\x1cgoogle/protobuf/struct.proto\x1a,rastervision/protos/raster_transformer.proto\x1a\'rastervision/protos/vector_source.proto\"\xaa\t\n\x12RasterSourceConfig\x12\x13\n\x0bsource_type\x18\x01 \x02(\t\x12\x38\n\x0ctransformers\x18\x02 \x03(\x0b\x32\".rv.protos.RasterTransformerConfig\x12\x15\n\rchannel_order\x18\x03 \x03(\x05\x12\x43\n\rgeotiff_files\x18\x04 \x01(\x0b\x32*.rv.protos.RasterSourceConfig.GeoTiffFilesH\x00\x12=\n\nimage_file\x18\x05 \x01(\x0b\x32\'.rv.protos.RasterSourceConfig.ImageFileH\x00\x12\x41\n\x0cgeojson_file\x18\x06 \x01(\x0b\x32).rv.protos.RasterSourceConfig.GeoJSONFileH\x00\x12\x30\n\rcustom_config\x18\x07 \x01(\x0b\x32\x17.google.protobuf.StructH\x00\x12K\n\x11rasterized_source\x18\x08 \x01(\x0b\x32..rv.protos.RasterSourceConfig.RasterizedSourceH\x00\x12G\n\x0frasterio_source\x18\t \x01(\x0b\x32,.rv.protos.RasterSourceConfig.RasterioSourceH\x00\x1aL\n\x0cGeoTiffFiles\x12\x0c\n\x04uris\x18\x01 \x03(\t\x12\x16\n\x0ex_shift_meters\x18\x02 \x01(\x02\x12\x16\n\x0ey_shift_meters\x18\x03 \x01(\x02\x1a\x18\n\tImageFile\x12\x0b\n\x03uri\x18\x01 \x02(\t\x1aN\n\x0eRasterioSource\x12\x0c\n\x04uris\x18\x01 \x03(\t\x12\x16\n\x0ex_shift_meters\x18\x02 \x01(\x02\x12\x16\n\x0ey_shift_meters\x18\x03 \x01(\x02\x1a\x8d\x02\n\x10RasterizedSource\x12\x34\n\rvector_source\x18\x01 \x02(\x0b\x32\x1d.rv.protos.VectorSourceConfig\x12\\\n\x12rasterizer_options\x18\x02 \x02(\x0b\x32@.rv.protos.RasterSourceConfig.RasterizedSource.RasterizerOptions\x1a\x65\n\x11RasterizerOptions\x12\x1b\n\x13\x62\x61\x63kground_class_id\x18\x02 \x02(\x05\x12\x17\n\x0bline_buffer\x18\x03 \x01(\x05:\x02\x31\x35\x12\x1a\n\x0b\x61ll_touched\x18\x04 \x01(\x08:\x05\x66\x61lse\x1a\xbe\x01\n\x0bGeoJSONFile\x12\x0b\n\x03uri\x18\x01 \x02(\t\x12W\n\x12rasterizer_options\x18\x02 \x02(\x0b\x32;.rv.protos.RasterSourceConfig.GeoJSONFile.RasterizerOptions\x1aI\n\x11RasterizerOptions\x12\x1b\n\x13\x62\x61\x63kground_class_id\x18\x02 \x02(\x05\x12\x17\n\x0bline_buffer\x18\x03 \x01(\x05:\x02\x31\x35\x42\x16\n\x14raster_source_config')
  ,
  dependencies=[google_dot_protobuf_dot_struct__pb2.DESCRIPTOR,rastervision_dot_protos_dot_raster__transformer__pb2.DESCRIPTOR,rastervision_dot_protos_dot_vector__source__pb2.DESCRIPTOR,])
_sym_db.RegisterFileDescriptor(DESCRIPTOR)




_RASTERSOURCECONFIG_GEOTIFFFILES = _descriptor.Descriptor(
  name='GeoTiffFiles',
  full_name='rv.protos.RasterSourceConfig.GeoTiffFiles',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='uris', full_name='rv.protos.RasterSourceConfig.GeoTiffFiles.uris', index=0,
      number=1, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='x_shift_meters', full_name='rv.protos.RasterSourceConfig.GeoTiffFiles.x_shift_meters', index=1,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='y_shift_meters', full_name='rv.protos.RasterSourceConfig.GeoTiffFiles.y_shift_meters', index=2,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=695,
  serialized_end=771,
)

_RASTERSOURCECONFIG_IMAGEFILE = _descriptor.Descriptor(
  name='ImageFile',
  full_name='rv.protos.RasterSourceConfig.ImageFile',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='uri', full_name='rv.protos.RasterSourceConfig.ImageFile.uri', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=773,
  serialized_end=797,
)

_RASTERSOURCECONFIG_RASTERIOSOURCE = _descriptor.Descriptor(
  name='RasterioSource',
  full_name='rv.protos.RasterSourceConfig.RasterioSource',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='uris', full_name='rv.protos.RasterSourceConfig.RasterioSource.uris', index=0,
      number=1, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='x_shift_meters', full_name='rv.protos.RasterSourceConfig.RasterioSource.x_shift_meters', index=1,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='y_shift_meters', full_name='rv.protos.RasterSourceConfig.RasterioSource.y_shift_meters', index=2,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=799,
  serialized_end=877,
)

_RASTERSOURCECONFIG_RASTERIZEDSOURCE_RASTERIZEROPTIONS = _descriptor.Descriptor(
  name='RasterizerOptions',
  full_name='rv.protos.RasterSourceConfig.RasterizedSource.RasterizerOptions',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='background_class_id', full_name='rv.protos.RasterSourceConfig.RasterizedSource.RasterizerOptions.background_class_id', index=0,
      number=2, type=5, cpp_type=1, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='line_buffer', full_name='rv.protos.RasterSourceConfig.RasterizedSource.RasterizerOptions.line_buffer', index=1,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=15,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='all_touched', full_name='rv.protos.RasterSourceConfig.RasterizedSource.RasterizerOptions.all_touched', index=2,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1048,
  serialized_end=1149,
)

_RASTERSOURCECONFIG_RASTERIZEDSOURCE = _descriptor.Descriptor(
  name='RasterizedSource',
  full_name='rv.protos.RasterSourceConfig.RasterizedSource',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='vector_source', full_name='rv.protos.RasterSourceConfig.RasterizedSource.vector_source', index=0,
      number=1, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='rasterizer_options', full_name='rv.protos.RasterSourceConfig.RasterizedSource.rasterizer_options', index=1,
      number=2, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[_RASTERSOURCECONFIG_RASTERIZEDSOURCE_RASTERIZEROPTIONS, ],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=880,
  serialized_end=1149,
)

_RASTERSOURCECONFIG_GEOJSONFILE_RASTERIZEROPTIONS = _descriptor.Descriptor(
  name='RasterizerOptions',
  full_name='rv.protos.RasterSourceConfig.GeoJSONFile.RasterizerOptions',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='background_class_id', full_name='rv.protos.RasterSourceConfig.GeoJSONFile.RasterizerOptions.background_class_id', index=0,
      number=2, type=5, cpp_type=1, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='line_buffer', full_name='rv.protos.RasterSourceConfig.GeoJSONFile.RasterizerOptions.line_buffer', index=1,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=15,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1048,
  serialized_end=1121,
)

_RASTERSOURCECONFIG_GEOJSONFILE = _descriptor.Descriptor(
  name='GeoJSONFile',
  full_name='rv.protos.RasterSourceConfig.GeoJSONFile',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='uri', full_name='rv.protos.RasterSourceConfig.GeoJSONFile.uri', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='rasterizer_options', full_name='rv.protos.RasterSourceConfig.GeoJSONFile.rasterizer_options', index=1,
      number=2, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[_RASTERSOURCECONFIG_GEOJSONFILE_RASTERIZEROPTIONS, ],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1152,
  serialized_end=1342,
)

_RASTERSOURCECONFIG = _descriptor.Descriptor(
  name='RasterSourceConfig',
  full_name='rv.protos.RasterSourceConfig',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='source_type', full_name='rv.protos.RasterSourceConfig.source_type', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='transformers', full_name='rv.protos.RasterSourceConfig.transformers', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='channel_order', full_name='rv.protos.RasterSourceConfig.channel_order', index=2,
      number=3, type=5, cpp_type=1, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='geotiff_files', full_name='rv.protos.RasterSourceConfig.geotiff_files', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='image_file', full_name='rv.protos.RasterSourceConfig.image_file', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='geojson_file', full_name='rv.protos.RasterSourceConfig.geojson_file', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='custom_config', full_name='rv.protos.RasterSourceConfig.custom_config', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='rasterized_source', full_name='rv.protos.RasterSourceConfig.rasterized_source', index=7,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='rasterio_source', full_name='rv.protos.RasterSourceConfig.rasterio_source', index=8,
      number=9, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[_RASTERSOURCECONFIG_GEOTIFFFILES, _RASTERSOURCECONFIG_IMAGEFILE, _RASTERSOURCECONFIG_RASTERIOSOURCE, _RASTERSOURCECONFIG_RASTERIZEDSOURCE, _RASTERSOURCECONFIG_GEOJSONFILE, ],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='raster_source_config', full_name='rv.protos.RasterSourceConfig.raster_source_config',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=172,
  serialized_end=1366,
)

_RASTERSOURCECONFIG_GEOTIFFFILES.containing_type = _RASTERSOURCECONFIG
_RASTERSOURCECONFIG_IMAGEFILE.containing_type = _RASTERSOURCECONFIG
_RASTERSOURCECONFIG_RASTERIOSOURCE.containing_type = _RASTERSOURCECONFIG
_RASTERSOURCECONFIG_RASTERIZEDSOURCE_RASTERIZEROPTIONS.containing_type = _RASTERSOURCECONFIG_RASTERIZEDSOURCE
_RASTERSOURCECONFIG_RASTERIZEDSOURCE.fields_by_name['vector_source'].message_type = rastervision_dot_protos_dot_vector__source__pb2._VECTORSOURCECONFIG
_RASTERSOURCECONFIG_RASTERIZEDSOURCE.fields_by_name['rasterizer_options'].message_type = _RASTERSOURCECONFIG_RASTERIZEDSOURCE_RASTERIZEROPTIONS
_RASTERSOURCECONFIG_RASTERIZEDSOURCE.containing_type = _RASTERSOURCECONFIG
_RASTERSOURCECONFIG_GEOJSONFILE_RASTERIZEROPTIONS.containing_type = _RASTERSOURCECONFIG_GEOJSONFILE
_RASTERSOURCECONFIG_GEOJSONFILE.fields_by_name['rasterizer_options'].message_type = _RASTERSOURCECONFIG_GEOJSONFILE_RASTERIZEROPTIONS
_RASTERSOURCECONFIG_GEOJSONFILE.containing_type = _RASTERSOURCECONFIG
_RASTERSOURCECONFIG.fields_by_name['transformers'].message_type = rastervision_dot_protos_dot_raster__transformer__pb2._RASTERTRANSFORMERCONFIG
_RASTERSOURCECONFIG.fields_by_name['geotiff_files'].message_type = _RASTERSOURCECONFIG_GEOTIFFFILES
_RASTERSOURCECONFIG.fields_by_name['image_file'].message_type = _RASTERSOURCECONFIG_IMAGEFILE
_RASTERSOURCECONFIG.fields_by_name['geojson_file'].message_type = _RASTERSOURCECONFIG_GEOJSONFILE
_RASTERSOURCECONFIG.fields_by_name['custom_config'].message_type = google_dot_protobuf_dot_struct__pb2._STRUCT
_RASTERSOURCECONFIG.fields_by_name['rasterized_source'].message_type = _RASTERSOURCECONFIG_RASTERIZEDSOURCE
_RASTERSOURCECONFIG.fields_by_name['rasterio_source'].message_type = _RASTERSOURCECONFIG_RASTERIOSOURCE
_RASTERSOURCECONFIG.oneofs_by_name['raster_source_config'].fields.append(
  _RASTERSOURCECONFIG.fields_by_name['geotiff_files'])
_RASTERSOURCECONFIG.fields_by_name['geotiff_files'].containing_oneof = _RASTERSOURCECONFIG.oneofs_by_name['raster_source_config']
_RASTERSOURCECONFIG.oneofs_by_name['raster_source_config'].fields.append(
  _RASTERSOURCECONFIG.fields_by_name['image_file'])
_RASTERSOURCECONFIG.fields_by_name['image_file'].containing_oneof = _RASTERSOURCECONFIG.oneofs_by_name['raster_source_config']
_RASTERSOURCECONFIG.oneofs_by_name['raster_source_config'].fields.append(
  _RASTERSOURCECONFIG.fields_by_name['geojson_file'])
_RASTERSOURCECONFIG.fields_by_name['geojson_file'].containing_oneof = _RASTERSOURCECONFIG.oneofs_by_name['raster_source_config']
_RASTERSOURCECONFIG.oneofs_by_name['raster_source_config'].fields.append(
  _RASTERSOURCECONFIG.fields_by_name['custom_config'])
_RASTERSOURCECONFIG.fields_by_name['custom_config'].containing_oneof = _RASTERSOURCECONFIG.oneofs_by_name['raster_source_config']
_RASTERSOURCECONFIG.oneofs_by_name['raster_source_config'].fields.append(
  _RASTERSOURCECONFIG.fields_by_name['rasterized_source'])
_RASTERSOURCECONFIG.fields_by_name['rasterized_source'].containing_oneof = _RASTERSOURCECONFIG.oneofs_by_name['raster_source_config']
_RASTERSOURCECONFIG.oneofs_by_name['raster_source_config'].fields.append(
  _RASTERSOURCECONFIG.fields_by_name['rasterio_source'])
_RASTERSOURCECONFIG.fields_by_name['rasterio_source'].containing_oneof = _RASTERSOURCECONFIG.oneofs_by_name['raster_source_config']
DESCRIPTOR.message_types_by_name['RasterSourceConfig'] = _RASTERSOURCECONFIG

RasterSourceConfig = _reflection.GeneratedProtocolMessageType('RasterSourceConfig', (_message.Message,), dict(

  GeoTiffFiles = _reflection.GeneratedProtocolMessageType('GeoTiffFiles', (_message.Message,), dict(
    DESCRIPTOR = _RASTERSOURCECONFIG_GEOTIFFFILES,
    __module__ = 'rastervision.protos.raster_source_pb2'
    # @@protoc_insertion_point(class_scope:rv.protos.RasterSourceConfig.GeoTiffFiles)
    ))
  ,

  ImageFile = _reflection.GeneratedProtocolMessageType('ImageFile', (_message.Message,), dict(
    DESCRIPTOR = _RASTERSOURCECONFIG_IMAGEFILE,
    __module__ = 'rastervision.protos.raster_source_pb2'
    # @@protoc_insertion_point(class_scope:rv.protos.RasterSourceConfig.ImageFile)
    ))
  ,

  RasterioSource = _reflection.GeneratedProtocolMessageType('RasterioSource', (_message.Message,), dict(
    DESCRIPTOR = _RASTERSOURCECONFIG_RASTERIOSOURCE,
    __module__ = 'rastervision.protos.raster_source_pb2'
    # @@protoc_insertion_point(class_scope:rv.protos.RasterSourceConfig.RasterioSource)
    ))
  ,

  RasterizedSource = _reflection.GeneratedProtocolMessageType('RasterizedSource', (_message.Message,), dict(

    RasterizerOptions = _reflection.GeneratedProtocolMessageType('RasterizerOptions', (_message.Message,), dict(
      DESCRIPTOR = _RASTERSOURCECONFIG_RASTERIZEDSOURCE_RASTERIZEROPTIONS,
      __module__ = 'rastervision.protos.raster_source_pb2'
      # @@protoc_insertion_point(class_scope:rv.protos.RasterSourceConfig.RasterizedSource.RasterizerOptions)
      ))
    ,
    DESCRIPTOR = _RASTERSOURCECONFIG_RASTERIZEDSOURCE,
    __module__ = 'rastervision.protos.raster_source_pb2'
    # @@protoc_insertion_point(class_scope:rv.protos.RasterSourceConfig.RasterizedSource)
    ))
  ,

  GeoJSONFile = _reflection.GeneratedProtocolMessageType('GeoJSONFile', (_message.Message,), dict(

    RasterizerOptions = _reflection.GeneratedProtocolMessageType('RasterizerOptions', (_message.Message,), dict(
      DESCRIPTOR = _RASTERSOURCECONFIG_GEOJSONFILE_RASTERIZEROPTIONS,
      __module__ = 'rastervision.protos.raster_source_pb2'
      # @@protoc_insertion_point(class_scope:rv.protos.RasterSourceConfig.GeoJSONFile.RasterizerOptions)
      ))
    ,
    DESCRIPTOR = _RASTERSOURCECONFIG_GEOJSONFILE,
    __module__ = 'rastervision.protos.raster_source_pb2'
    # @@protoc_insertion_point(class_scope:rv.protos.RasterSourceConfig.GeoJSONFile)
    ))
  ,
  DESCRIPTOR = _RASTERSOURCECONFIG,
  __module__ = 'rastervision.protos.raster_source_pb2'
  # @@protoc_insertion_point(class_scope:rv.protos.RasterSourceConfig)
  ))
_sym_db.RegisterMessage(RasterSourceConfig)
_sym_db.RegisterMessage(RasterSourceConfig.GeoTiffFiles)
_sym_db.RegisterMessage(RasterSourceConfig.ImageFile)
_sym_db.RegisterMessage(RasterSourceConfig.RasterioSource)
_sym_db.RegisterMessage(RasterSourceConfig.RasterizedSource)
_sym_db.RegisterMessage(RasterSourceConfig.RasterizedSource.RasterizerOptions)
_sym_db.RegisterMessage(RasterSourceConfig.GeoJSONFile)
_sym_db.RegisterMessage(RasterSourceConfig.GeoJSONFile.RasterizerOptions)


# @@protoc_insertion_point(module_scope)



from copy import deepcopy

import rastervision as rv
from rastervision.command import (EvalCommand, CommandConfig,
                                  CommandConfigBuilder, NoOpCommand)
from rastervision.protos.command_pb2 \
    import CommandConfig as CommandConfigMsg
from rastervision.rv_config import RVConfig
from rastervision.command.utils import (check_scenes_type, check_task_type)
from rastervision.evaluation import EvaluatorConfig


class EvalCommandConfig(CommandConfig):
    def __init__(self, root_uri, task, scenes, evaluators):
        super().__init__(rv.EVAL, root_uri)
        self.task = task
        self.scenes = scenes
        self.evaluators = evaluators

    def create_command(self, tmp_dir=None):
        if len(self.scenes) == 0 or len(self.evaluators) == 0:
            return NoOpCommand()

        if not tmp_dir:
            _tmp_dir = RVConfig.get_tmp_dir()
            tmp_dir = _tmp_dir.name
        else:
            _tmp_dir = tmp_dir

        retval = EvalCommand(self)
        retval.set_tmp_dir(_tmp_dir)
        return retval

    def to_proto(self):
        msg = super().to_proto()
        task = self.task.to_proto()
        scenes = list(map(lambda s: s.to_proto(), self.scenes))
        evaluators = list(map(lambda e: e.to_proto(), self.evaluators))

        msg.MergeFrom(
            CommandConfigMsg(
                eval_config=CommandConfigMsg.EvalConfig(
                    task=task, scenes=scenes, evaluators=evaluators)))

        return msg

    def report_io(self):
        io_def = rv.core.CommandIODefinition()
        self.task.report_io(self.command_type, io_def)
        for scene in self.scenes:
            scene.report_io(self.command_type, io_def)
        for evaluator in self.evaluators:
            evaluator.report_io(self.command_type, io_def)
        return io_def


class EvalCommandConfigBuilder(CommandConfigBuilder):
    def __init__(self, command_type, prev=None):
        super().__init__(command_type, prev)
        if prev is None:
            self.task = None
            self.scenes = None
            self.evaluators = None
        else:
            self.task = prev.task
            self.scenes = prev.scenes
            self.evaluators = prev.evaluators

    def validate(self):
        super().validate()
        if self.task is None:
            raise rv.ConfigError(
                'task not set for EvalCommandConfig. Use with_task or '
                'with_experiment')
        check_task_type(self.task)
        if self.scenes is None:
            raise rv.ConfigError(
                'scenes not set for EvalCommandConfig. Use with_scenes or '
                'with_experiment')
        check_scenes_type(self.scenes)
        if self.evaluators is None:
            raise rv.ConfigError(
                'evaluators not set. Use with_evaluators or with_experiment')
        if not isinstance(self.evaluators, list):
            raise rv.ConfigError(
                'evaluators must be a list of EvaluatorConfig objects, got {}'.
                format(type(self.evaluators)))
        for evaluator in self.evaluators:
            if not issubclass(type(evaluator), EvaluatorConfig):
                if not isinstance(evaluator, str):
                    raise rv.ConfigError(
                        'evaluators must be a subclass of EvaluatorConfig or string,'
                        ' got {}'.format(type(evaluator)))

    def build(self):
        self.validate()
        return EvalCommandConfig(self.root_uri, self.task, self.scenes,
                                 self.evaluators)

    def from_proto(self, msg):
        b = super().from_proto(msg)

        conf = msg.eval_config

        task = rv.TaskConfig.from_proto(conf.task)
        scenes = list(map(rv.SceneConfig.from_proto, conf.scenes))
        evaluators = list(map(rv.EvaluatorConfig.from_proto, conf.evaluators))

        b = b.with_task(task)
        b = b.with_scenes(scenes)
        b = b.with_evaluators(evaluators)

        return b

    def get_root_uri(self, experiment_config):
        return experiment_config.eval_uri

    def with_experiment(self, experiment_config):
        b = super().with_experiment(experiment_config)
        b = b.with_task(experiment_config.task)
        b = b.with_scenes(experiment_config.dataset.validation_scenes)
        b = b.with_evaluators(experiment_config.evaluators)
        return b

    def with_task(self, task):
        """Sets a specific task type.

        Args:
            task:  A TaskConfig object.

        """
        b = deepcopy(self)
        b.task = task
        return b

    def with_scenes(self, scenes):
        b = deepcopy(self)
        b.scenes = scenes
        return b

    def with_evaluators(self, evaluators):
        b = deepcopy(self)
        b.evaluators = evaluators
        return b
